"""
Comprehensive tests for SAML auth modules to maximize coverage.

Targets uncovered lines in:
- saml_handler.py (lines 60-62, 90-95, 319-322, 352-366, 432-435, 455-458,
                   503-523, 576-632, 645-692, 750-798, 817-848, 904-929)
- saml_config.py (lines 134, 136, 144, 146, 160, 175-186, 197-198, 210-211,
                  230-246, 298, 305, 309, 335, 346, 355, 406, 414, 422-429,
                  441-482)
- saml_types.py (lines 97-180)
- provisioning.py (lines 277, 330-456, 551-582, 601-615)
- role_mapper.py (lines 194, 313-318, 333-344, 351, 355, 392-405, 478-502,
                  518-539, 555-568)
"""

from __future__ import annotations

import os
import sys
from datetime import UTC, datetime
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

# ---------------------------------------------------------------------------
# saml_config.py tests
# ---------------------------------------------------------------------------


class TestSAMLSPConfigPostInit:
    """Cover __post_init__ validation in SAMLSPConfig (lines 134, 136)."""

    def test_empty_entity_id_raises(self):
        from src.core.shared.auth.saml_config import SAMLConfigurationError, SAMLSPConfig

        with pytest.raises(SAMLConfigurationError, match="entity ID"):
            SAMLSPConfig(entity_id="", acs_url="http://acs")

    def test_empty_acs_url_raises(self):
        from src.core.shared.auth.saml_config import SAMLConfigurationError, SAMLSPConfig

        with pytest.raises(SAMLConfigurationError, match="ACS URL"):
            SAMLSPConfig(entity_id="urn:test", acs_url="")


class TestSAMLSPConfigDefaultCerts:
    """Cover lines 143-146: default cert/key file usage when files exist."""

    def test_default_cert_used_when_file_exists(self, tmp_path):
        from src.core.shared.auth.saml_config import SAMLSPConfig

        cert = tmp_path / "sp.crt"
        cert.write_text("CERT")
        key = tmp_path / "sp.key"
        key.write_text("KEY")

        with (
            patch("src.core.shared.auth.saml_config.DEFAULT_SP_CERT", cert),
            patch("src.core.shared.auth.saml_config.DEFAULT_SP_KEY", key),
        ):
            cfg = SAMLSPConfig(entity_id="urn:test", acs_url="/acs")
            assert cfg.cert_file == str(cert)
            assert cfg.key_file == str(key)


class TestSAMLSPConfigFindXmlsec:
    """Cover _find_xmlsec_binary lines 160, 175-186."""

    def test_env_var_valid(self, tmp_path):
        from src.core.shared.auth.saml_config import SAMLSPConfig

        fake = tmp_path / "xmlsec1"
        fake.write_text("#!/bin/sh")
        with patch.dict(os.environ, {"SAML_XMLSEC_BINARY": str(fake)}):
            cfg = SAMLSPConfig(entity_id="urn:x", acs_url="/acs")
            assert cfg.xmlsec_binary == str(fake)

    def test_env_var_not_file(self):
        from src.core.shared.auth.saml_config import SAMLSPConfig

        with patch.dict(os.environ, {"SAML_XMLSEC_BINARY": "/nonexistent/xmlsec1"}):
            # Falls through to common paths and shutil.which
            with patch("shutil.which", return_value=None):
                with patch("os.path.isfile", side_effect=lambda p: False):
                    cfg = SAMLSPConfig(entity_id="urn:x", acs_url="/acs")
                    assert cfg.xmlsec_binary is None

    def test_shutil_which_fallback(self):
        from src.core.shared.auth.saml_config import SAMLSPConfig

        with (
            patch.dict(os.environ, {}, clear=False),
            patch("os.path.isfile", return_value=False),
            patch("shutil.which", return_value="/usr/bin/xmlsec1"),
        ):
            os.environ.pop("SAML_XMLSEC_BINARY", None)
            cfg = SAMLSPConfig(entity_id="urn:x", acs_url="/acs")
            assert cfg.xmlsec_binary == "/usr/bin/xmlsec1"

    def test_nothing_found_returns_none(self):
        from src.core.shared.auth.saml_config import SAMLSPConfig

        with (
            patch.dict(os.environ, {}, clear=False),
            patch("os.path.isfile", return_value=False),
            patch("shutil.which", return_value=None),
        ):
            os.environ.pop("SAML_XMLSEC_BINARY", None)
            cfg = SAMLSPConfig(entity_id="urn:x", acs_url="/acs")
            assert cfg.xmlsec_binary is None


class TestSAMLSPConfigGetCertContent:
    """Cover get_cert_content / get_key_content (lines 197-198, 210-211)."""

    def test_cert_content_from_file(self, tmp_path):
        from src.core.shared.auth.saml_config import SAMLSPConfig

        cert = tmp_path / "c.crt"
        cert.write_text("CERTDATA")
        cfg = SAMLSPConfig(entity_id="urn:x", acs_url="/a", cert_file=str(cert))
        assert cfg.get_cert_content() == "CERTDATA"

    def test_key_content_from_file(self, tmp_path):
        from src.core.shared.auth.saml_config import SAMLSPConfig

        key = tmp_path / "k.key"
        key.write_text("KEYDATA")
        cfg = SAMLSPConfig(entity_id="urn:x", acs_url="/a", key_file=str(key))
        assert cfg.get_key_content() == "KEYDATA"

    def test_cert_content_inline_takes_precedence(self, tmp_path):
        from src.core.shared.auth.saml_config import SAMLSPConfig

        cert = tmp_path / "c.crt"
        cert.write_text("FILE")
        cfg = SAMLSPConfig(
            entity_id="urn:x", acs_url="/a", cert_file=str(cert), cert_content="INLINE"
        )
        assert cfg.get_cert_content() == "INLINE"

    def test_key_content_inline_takes_precedence(self, tmp_path):
        from src.core.shared.auth.saml_config import SAMLSPConfig

        key = tmp_path / "k.key"
        key.write_text("FILE")
        cfg = SAMLSPConfig(entity_id="urn:x", acs_url="/a", key_file=str(key), key_content="INLINE")
        assert cfg.get_key_content() == "INLINE"


class TestSAMLSPConfigValidate:
    """Cover validate() lines 230-246."""

    def test_validate_sign_without_creds(self):
        from src.core.shared.auth.saml_config import SAMLSPConfig

        cfg = SAMLSPConfig(entity_id="urn:x", acs_url="/a", sign_authn_requests=True)
        errors = cfg.validate()
        assert any("certificate and key" in e for e in errors)

    def test_validate_sign_without_xmlsec(self):
        from src.core.shared.auth.saml_config import SAMLSPConfig

        with (
            patch("os.path.isfile", return_value=False),
            patch("shutil.which", return_value=None),
        ):
            os.environ.pop("SAML_XMLSEC_BINARY", None)
            cfg = SAMLSPConfig(
                entity_id="urn:x",
                acs_url="/a",
                sign_authn_requests=True,
                cert_content="CERT",
                key_content="KEY",
            )
            assert cfg.xmlsec_binary is None
            errors = cfg.validate()
            assert any("xmlsec1 binary is required" in e for e in errors)

    def test_validate_sign_xmlsec_not_found(self, tmp_path):
        from src.core.shared.auth.saml_config import SAMLSPConfig

        cfg = SAMLSPConfig(
            entity_id="urn:x",
            acs_url="/a",
            sign_authn_requests=True,
            cert_content="CERT",
            key_content="KEY",
            xmlsec_binary="/nonexistent/xmlsec1",
        )
        errors = cfg.validate()
        assert any("not found at" in e for e in errors)

    def test_validate_no_sign_no_errors(self):
        from src.core.shared.auth.saml_config import SAMLSPConfig

        cfg = SAMLSPConfig(entity_id="urn:x", acs_url="/a", sign_authn_requests=False)
        errors = cfg.validate()
        assert not errors


class TestSAMLIdPConfigPostInit:
    """Cover IdP __post_init__ lines 298, 305, 309."""

    def test_empty_name_raises(self):
        from src.core.shared.auth.saml_config import SAMLConfigurationError, SAMLIdPConfig

        with pytest.raises(SAMLConfigurationError, match="name is required"):
            SAMLIdPConfig(name="")

    def test_invalid_sso_binding_raises(self):
        from src.core.shared.auth.saml_config import SAMLConfigurationError, SAMLIdPConfig

        with pytest.raises(SAMLConfigurationError, match="Invalid SSO binding"):
            SAMLIdPConfig(name="test", sso_binding="invalid")

    def test_invalid_slo_binding_raises(self):
        from src.core.shared.auth.saml_config import SAMLConfigurationError, SAMLIdPConfig

        with pytest.raises(SAMLConfigurationError, match="Invalid SLO binding"):
            SAMLIdPConfig(name="test", slo_binding="invalid")


class TestSAMLIdPConfigMethods:
    """Cover IdP helper methods lines 335, 346, 355."""

    def test_is_configured_with_metadata_url(self):
        from src.core.shared.auth.saml_config import SAMLIdPConfig

        idp = SAMLIdPConfig(name="test", metadata_url="http://meta")
        assert idp.is_configured()

    def test_is_configured_with_manual(self):
        from src.core.shared.auth.saml_config import SAMLIdPConfig

        idp = SAMLIdPConfig(name="test", sso_url="http://sso", certificate="CERT")
        assert idp.is_configured()

    def test_not_configured(self):
        from src.core.shared.auth.saml_config import SAMLIdPConfig

        idp = SAMLIdPConfig(name="test")
        assert not idp.is_configured()

    def test_validate_no_config(self):
        from src.core.shared.auth.saml_config import SAMLIdPConfig

        idp = SAMLIdPConfig(name="test")
        errors = idp.validate()
        assert any("metadata" in e.lower() or "manual" in e.lower() for e in errors)

    def test_validate_manual_no_entity_id(self):
        from src.core.shared.auth.saml_config import SAMLIdPConfig

        idp = SAMLIdPConfig(name="test", sso_url="http://sso", certificate="CERT")
        errors = idp.validate()
        assert any("entity ID" in e for e in errors)

    def test_validate_ok_with_metadata(self):
        from src.core.shared.auth.saml_config import SAMLIdPConfig

        idp = SAMLIdPConfig(name="test", metadata_url="http://meta")
        errors = idp.validate()
        assert not errors


class TestSAMLConfigMethods:
    """Cover SAMLConfig lines 406, 414, 422-429."""

    def test_get_idp(self):
        from src.core.shared.auth.saml_config import SAMLConfig, SAMLIdPConfig, SAMLSPConfig

        sp = SAMLSPConfig(entity_id="urn:x", acs_url="/a")
        cfg = SAMLConfig(sp=sp)
        idp = SAMLIdPConfig(name="okta", metadata_url="http://m")
        cfg.add_idp(idp)
        assert cfg.get_idp("okta") is idp
        assert cfg.get_idp("nonexistent") is None

    def test_list_enabled_idps(self):
        from src.core.shared.auth.saml_config import SAMLConfig, SAMLIdPConfig, SAMLSPConfig

        sp = SAMLSPConfig(entity_id="urn:x", acs_url="/a")
        cfg = SAMLConfig(sp=sp)
        cfg.add_idp(SAMLIdPConfig(name="a", metadata_url="http://m", enabled=True))
        cfg.add_idp(SAMLIdPConfig(name="b", metadata_url="http://m", enabled=False))
        assert cfg.list_enabled_idps() == ["a"]

    def test_validate_aggregates_errors(self):
        from src.core.shared.auth.saml_config import SAMLConfig, SAMLIdPConfig, SAMLSPConfig

        sp = SAMLSPConfig(entity_id="urn:x", acs_url="/a", sign_authn_requests=False)
        cfg = SAMLConfig(sp=sp)
        # IdP with no config
        cfg.add_idp(SAMLIdPConfig(name="bad"))
        errors = cfg.validate()
        assert any("IdP 'bad'" in e for e in errors)


class TestSAMLConfigFromSettings:
    """Cover from_settings lines 441-482."""

    def test_from_settings_no_sso(self):
        from src.core.shared.auth.saml_config import SAMLConfig, SAMLConfigurationError

        settings = MagicMock(spec=[])
        with pytest.raises(SAMLConfigurationError, match="SSO settings not found"):
            SAMLConfig.from_settings(settings)

    def test_from_settings_basic(self):
        from src.core.shared.auth.saml_config import SAMLConfig

        sso = MagicMock()
        sso.saml_entity_id = "urn:test"
        sso.saml_sign_requests = False
        sso.saml_want_assertions_signed = True
        sso.saml_want_assertions_encrypted = False
        sso.saml_sp_certificate = "CERT"
        sso.saml_sp_private_key = "KEY"
        sso.saml_idp_metadata_url = "http://idp/meta"
        sso.saml_idp_sso_url = "http://idp/sso"
        sso.saml_idp_slo_url = "http://idp/slo"
        sso.saml_idp_certificate = "IDP_CERT"

        settings = MagicMock()
        settings.sso = sso

        cfg = SAMLConfig.from_settings(settings)
        assert cfg.sp.entity_id == "urn:test"
        assert cfg.sp.cert_content == "CERT"
        assert cfg.sp.key_content == "KEY"
        assert "default" in cfg.idps

    def test_from_settings_secret_key(self):
        from src.core.shared.auth.saml_config import SAMLConfig

        secret_key = MagicMock()
        secret_key.get_secret_value.return_value = "SECRET_KEY"

        sso = MagicMock()
        sso.saml_entity_id = "urn:t"
        sso.saml_sign_requests = False
        sso.saml_want_assertions_signed = True
        sso.saml_want_assertions_encrypted = False
        sso.saml_sp_certificate = None
        sso.saml_sp_private_key = secret_key
        sso.saml_idp_metadata_url = None

        settings = MagicMock()
        settings.sso = sso

        cfg = SAMLConfig.from_settings(settings)
        assert cfg.sp.key_content == "SECRET_KEY"

    def test_from_settings_no_idp_metadata(self):
        from src.core.shared.auth.saml_config import SAMLConfig

        sso = MagicMock()
        sso.saml_entity_id = "urn:t"
        sso.saml_sign_requests = False
        sso.saml_want_assertions_signed = True
        sso.saml_want_assertions_encrypted = False
        sso.saml_sp_certificate = None
        sso.saml_sp_private_key = None
        sso.saml_idp_metadata_url = None

        settings = MagicMock()
        settings.sso = sso

        cfg = SAMLConfig.from_settings(settings)
        assert len(cfg.idps) == 0


# ---------------------------------------------------------------------------
# saml_types.py tests
# ---------------------------------------------------------------------------


class TestSAMLUserInfoFromResponse:
    """Cover SAMLUserInfo.from_response lines 97-180."""

    def test_no_pysaml2_raises(self):
        from src.core.shared.auth.saml_types import SAMLUserInfo

        with pytest.raises(Exception, match="PySAML2"):
            SAMLUserInfo.from_response(MagicMock(), has_pysaml2=False)

    def test_full_attribute_extraction(self):
        from src.core.shared.auth.saml_types import SAMLUserInfo

        name_id = MagicMock()
        name_id.__str__ = lambda self: "user@example.com"
        name_id.format = "urn:oasis:names:tc:SAML:1.1:nameid-format:emailAddress"

        response = MagicMock()
        response.name_id = name_id
        response.session_info.return_value = {"session_index": "_sess123"}
        response.ava = {
            "email": ["user@example.com"],
            "name": ["John Doe"],
            "givenName": ["John"],
            "surname": ["Doe"],
            "groups": ["Engineering", "Admin"],
        }
        response.issuer.return_value = "http://idp.example.com"

        info = SAMLUserInfo.from_response(response, has_pysaml2=True)
        assert info.name_id == "user@example.com"
        assert info.email == "user@example.com"
        assert info.name == "John Doe"
        assert info.given_name == "John"
        assert info.family_name == "Doe"
        assert info.groups == ["Engineering", "Admin"]
        assert info.issuer == "http://idp.example.com"
        assert info.session_index == "_sess123"

    def test_claims_uri_attributes(self):
        from src.core.shared.auth.saml_types import SAMLUserInfo

        name_id = MagicMock()
        name_id.__str__ = lambda self: "u@e.com"
        name_id.format = None

        response = MagicMock()
        response.name_id = name_id
        response.session_info.return_value = {}
        response.ava = {
            "http://schemas.xmlsoap.org/ws/2005/05/identity/claims/emailaddress": ["alt@e.com"],
            "http://schemas.xmlsoap.org/ws/2005/05/identity/claims/name": ["Alt Name"],
            "http://schemas.xmlsoap.org/ws/2005/05/identity/claims/givenname": ["Alt"],
            "http://schemas.xmlsoap.org/ws/2005/05/identity/claims/surname": ["Name"],
            "http://schemas.microsoft.com/ws/2008/06/identity/claims/groups": ["G1"],
        }
        response.issuer.return_value = None

        info = SAMLUserInfo.from_response(response, has_pysaml2=True)
        assert info.email == "alt@e.com"
        assert info.name == "Alt Name"
        assert info.given_name == "Alt"
        assert info.family_name == "Name"
        assert info.groups == ["G1"]

    def test_no_name_id(self):
        from src.core.shared.auth.saml_types import SAMLUserInfo

        response = MagicMock()
        response.name_id = None
        response.session_info.return_value = {}
        response.ava = {}
        response.issuer.return_value = None

        info = SAMLUserInfo.from_response(response, has_pysaml2=True)
        assert info.name_id == ""
        assert info.email == ""  # Falls back to empty name_id

    def test_empty_attr_lists(self):
        from src.core.shared.auth.saml_types import SAMLUserInfo

        name_id = MagicMock()
        name_id.__str__ = lambda self: "u"
        name_id.format = "fmt"

        response = MagicMock()
        response.name_id = name_id
        response.session_info.return_value = {}
        response.ava = {
            "email": [],
            "name": [],
            "givenName": [],
            "surname": [],
            "groups": [],
        }
        response.issuer.return_value = None

        info = SAMLUserInfo.from_response(response, has_pysaml2=True)
        assert info.email is None or info.email == "u"  # Falls back to name_id
        assert info.groups == []

    def test_no_issuer_method(self):
        from src.core.shared.auth.saml_types import SAMLUserInfo

        name_id = MagicMock()
        name_id.__str__ = lambda self: "u"
        name_id.format = None

        response = MagicMock(spec=[])
        response.name_id = name_id
        response.session_info = MagicMock(return_value={})
        response.ava = {}

        info = SAMLUserInfo.from_response(response, has_pysaml2=True)
        assert info.issuer is None

    def test_alt_email_attrs(self):
        """Cover emailAddress and mail attribute paths."""
        from src.core.shared.auth.saml_types import SAMLUserInfo

        name_id = MagicMock()
        name_id.__str__ = lambda self: "u"
        name_id.format = None

        response = MagicMock()
        response.name_id = name_id
        response.session_info.return_value = {}
        response.ava = {"mail": ["mail@example.com"]}
        response.issuer.return_value = None

        info = SAMLUserInfo.from_response(response, has_pysaml2=True)
        assert info.email == "mail@example.com"

    def test_alt_name_attrs(self):
        """Cover displayName, cn, firstName, lastName, familyName, memberOf."""
        from src.core.shared.auth.saml_types import SAMLUserInfo

        name_id = MagicMock()
        name_id.__str__ = lambda self: "u"
        name_id.format = None

        response = MagicMock()
        response.name_id = name_id
        response.session_info.return_value = {}
        response.ava = {
            "displayName": ["Display"],
            "firstName": ["First"],
            "lastName": ["Last"],
            "memberOf": ["Group1", "Group2"],
        }
        response.issuer.return_value = None

        info = SAMLUserInfo.from_response(response, has_pysaml2=True)
        assert info.name == "Display"
        assert info.given_name == "First"
        assert info.family_name == "Last"
        assert info.groups == ["Group1", "Group2"]

    def test_emailAddress_attr(self):
        """Cover emailAddress attribute key."""
        from src.core.shared.auth.saml_types import SAMLUserInfo

        name_id = MagicMock()
        name_id.__str__ = lambda self: "u"
        name_id.format = None

        response = MagicMock()
        response.name_id = name_id
        response.session_info.return_value = {}
        response.ava = {"emailAddress": ["ea@example.com"]}
        response.issuer.return_value = None

        info = SAMLUserInfo.from_response(response, has_pysaml2=True)
        assert info.email == "ea@example.com"

    def test_cn_attr(self):
        """Cover cn attribute key for name."""
        from src.core.shared.auth.saml_types import SAMLUserInfo

        name_id = MagicMock()
        name_id.__str__ = lambda self: "u"
        name_id.format = None

        response = MagicMock()
        response.name_id = name_id
        response.session_info.return_value = {}
        response.ava = {
            "cn": ["Common Name"],
            "sn": ["Surname"],
            "familyName": ["Family"],  # Won't be used since sn comes first
        }
        response.issuer.return_value = None

        info = SAMLUserInfo.from_response(response, has_pysaml2=True)
        assert info.name == "Common Name"
        assert info.family_name == "Surname"


# ---------------------------------------------------------------------------
# saml_handler.py tests
# ---------------------------------------------------------------------------


class TestSAMLHandlerInit:
    """Cover handler init and basic methods."""

    def test_default_init(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        assert handler.sp_config.entity_id == "urn:acgs2:saml:sp"

    def test_init_with_sp_config(self):
        from src.core.shared.auth.saml_config import SAMLSPConfig
        from src.core.shared.auth.saml_handler import SAMLHandler

        sp = SAMLSPConfig(entity_id="urn:custom", acs_url="/custom/acs")
        handler = SAMLHandler(sp_config=sp)
        assert handler.sp_config.entity_id == "urn:custom"

    def test_init_with_full_config(self):
        from src.core.shared.auth.saml_config import SAMLConfig, SAMLSPConfig
        from src.core.shared.auth.saml_handler import SAMLHandler

        sp = SAMLSPConfig(entity_id="urn:full", acs_url="/full/acs")
        config = SAMLConfig(sp=sp)
        handler = SAMLHandler(config=config)
        assert handler.sp_config.entity_id == "urn:full"


class TestSAMLHandlerRegisterIdP:
    """Cover register_idp and related."""

    def test_register_idp_with_metadata_url(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://meta")
        assert "okta" in handler.list_idps()
        assert handler.get_idp("okta").metadata_url == "http://meta"

    def test_register_idp_invalid_raises(self):
        from src.core.shared.auth.saml_config import SAMLConfigurationError
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        with pytest.raises(SAMLConfigurationError, match="Invalid IdP"):
            handler.register_idp(name="bad")

    def test_register_clears_cached_client(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        handler._saml_clients["okta"] = MagicMock()
        handler.register_idp(name="okta", metadata_url="http://meta")
        assert "okta" not in handler._saml_clients

    def test_get_idp_not_registered(self):
        from src.core.shared.auth.saml_config import SAMLConfigurationError
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        with pytest.raises(SAMLConfigurationError, match="not registered"):
            handler.get_idp("missing")


class TestSAMLHandlerRegisterIdPFromModel:
    """Cover register_idp_from_model line 277 (not SAML) path."""

    def test_not_saml_raises(self):
        from src.core.shared.auth.saml_config import SAMLConfigurationError
        from src.core.shared.auth.saml_handler import SAMLHandler

        provider = MagicMock()
        provider.is_saml = False
        provider.name = "oidc-provider"

        handler = SAMLHandler()
        with pytest.raises(SAMLConfigurationError, match="not a SAML"):
            handler.register_idp_from_model(provider)

    def test_saml_validation_errors(self):
        from src.core.shared.auth.saml_config import SAMLConfigurationError
        from src.core.shared.auth.saml_handler import SAMLHandler

        provider = MagicMock()
        provider.is_saml = True
        provider.name = "bad-saml"
        provider.validate_saml_config.return_value = ["missing cert"]

        handler = SAMLHandler()
        with pytest.raises(SAMLConfigurationError, match="Invalid SAML"):
            handler.register_idp_from_model(provider)


class TestSAMLHandlerHttpClient:
    """Cover _get_http_client lines 319-322."""

    @pytest.mark.asyncio
    async def test_get_http_client_no_httpx(self):
        from src.core.shared.auth.saml_handler import SAMLHandler
        from src.core.shared.auth.saml_types import SAMLError

        handler = SAMLHandler()
        with patch("src.core.shared.auth.saml_handler.HAS_HTTPX", False):
            with pytest.raises(SAMLError, match="httpx"):
                await handler._get_http_client()

    @pytest.mark.asyncio
    async def test_get_http_client_creates_client(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        mock_client = MagicMock()
        with (
            patch("src.core.shared.auth.saml_handler.HAS_HTTPX", True),
            patch("src.core.shared.auth.saml_handler.httpx") as mock_httpx,
        ):
            mock_httpx.AsyncClient.return_value = mock_client
            mock_httpx.Timeout = MagicMock()
            client = await handler._get_http_client()
            assert client is mock_client
            # Second call returns cached
            client2 = await handler._get_http_client()
            assert client2 is mock_client


class TestSAMLHandlerFetchMetadata:
    """Cover _fetch_metadata lines 352-366."""

    @pytest.mark.asyncio
    async def test_returns_inline_xml(self):
        from src.core.shared.auth.saml_config import SAMLIdPConfig
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        idp = SAMLIdPConfig(name="test", metadata_xml="<xml/>")
        result = await handler._fetch_metadata(idp)
        assert result == "<xml/>"

    @pytest.mark.asyncio
    async def test_returns_cached(self):
        from src.core.shared.auth.saml_config import SAMLIdPConfig
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        idp = SAMLIdPConfig(name="test", metadata_url="http://meta")
        handler._metadata_cache["test"] = ("<cached/>", datetime.now(UTC))
        result = await handler._fetch_metadata(idp)
        assert result == "<cached/>"

    @pytest.mark.asyncio
    async def test_fetch_success(self):
        from src.core.shared.auth.saml_config import SAMLIdPConfig
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        idp = SAMLIdPConfig(name="test", metadata_url="http://meta")

        mock_response = MagicMock()
        mock_response.text = "<fresh/>"
        mock_response.raise_for_status = MagicMock()

        mock_client = AsyncMock()
        mock_client.get.return_value = mock_response
        handler._http_client = mock_client

        with patch("src.core.shared.auth.saml_handler.HAS_HTTPX", True):
            result = await handler._fetch_metadata(idp, force_refresh=True)
            assert result == "<fresh/>"
            assert "test" in handler._metadata_cache

    @pytest.mark.asyncio
    async def test_fetch_error_uses_cache(self):
        from src.core.shared.auth.saml_config import SAMLIdPConfig
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        idp = SAMLIdPConfig(name="test", metadata_url="http://meta")
        handler._metadata_cache["test"] = ("<old/>", datetime.now(UTC))

        mock_client = AsyncMock()
        mock_client.get.side_effect = RuntimeError("network error")
        handler._http_client = mock_client

        with patch("src.core.shared.auth.saml_handler.HAS_HTTPX", True):
            result = await handler._fetch_metadata(idp, force_refresh=True)
            assert result == "<old/>"

    @pytest.mark.asyncio
    async def test_fetch_error_no_cache_raises(self):
        from src.core.shared.auth.saml_config import SAMLIdPConfig
        from src.core.shared.auth.saml_handler import SAMLHandler
        from src.core.shared.auth.saml_types import SAMLProviderError

        handler = SAMLHandler()
        idp = SAMLIdPConfig(name="test", metadata_url="http://meta")

        mock_client = AsyncMock()
        mock_client.get.side_effect = RuntimeError("network error")
        handler._http_client = mock_client

        with (
            patch("src.core.shared.auth.saml_handler.HAS_HTTPX", True),
            pytest.raises(SAMLProviderError, match="Failed to fetch"),
        ):
            await handler._fetch_metadata(idp, force_refresh=True)


class TestSAMLHandlerBuildConfig:
    """Cover _build_pysaml2_config lines 432-435, 455-458."""

    def test_build_with_certs(self, tmp_path):
        from src.core.shared.auth.saml_config import SAMLIdPConfig, SAMLSPConfig
        from src.core.shared.auth.saml_handler import SAMLHandler

        sp = SAMLSPConfig(
            entity_id="urn:x",
            acs_url="/acs",
            sls_url="/sls",
            cert_content="CERT",
            key_content="KEY",
        )
        handler = SAMLHandler(sp_config=sp)
        idp = SAMLIdPConfig(name="test", metadata_url="http://m")

        config = handler._build_pysaml2_config(idp, "<metadata/>")
        assert "cert_file" in config
        assert "key_file" in config
        assert "metadata" in config

    def test_build_manual_idp(self):
        from src.core.shared.auth.saml_config import SAMLIdPConfig, SAMLSPConfig
        from src.core.shared.auth.saml_handler import SAMLHandler

        sp = SAMLSPConfig(entity_id="urn:x", acs_url="/acs")
        handler = SAMLHandler(sp_config=sp)
        idp = SAMLIdPConfig(
            name="manual",
            entity_id="urn:idp",
            sso_url="http://sso",
            slo_url="http://slo",
            certificate="CERT",
        )

        config = handler._build_pysaml2_config(idp, None)
        assert config["metadata"]["inline"][0]["entity_id"] == "urn:idp"

    def test_build_cleanup_old_temp_files(self):
        from src.core.shared.auth.saml_config import SAMLIdPConfig, SAMLSPConfig
        from src.core.shared.auth.saml_handler import SAMLHandler

        sp = SAMLSPConfig(entity_id="urn:x", acs_url="/acs", cert_content="C", key_content="K")
        handler = SAMLHandler(sp_config=sp)
        idp = SAMLIdPConfig(name="t", metadata_url="http://m")

        # Set old temp files
        old_cert = MagicMock()
        old_cert.name = "/tmp/old.crt"
        old_key = MagicMock()
        old_key.name = "/tmp/old.key"
        old_meta = MagicMock()
        old_meta.name = "/tmp/old.xml"
        handler._temp_cert_file = old_cert
        handler._temp_key_file = old_key
        handler._temp_metadata_file = old_meta

        with patch("pathlib.Path.unlink"):
            config = handler._build_pysaml2_config(idp, "<xml/>")
            assert "cert_file" in config


class TestSAMLHandlerGetSamlClient:
    """Cover _get_saml_client lines 503-523."""

    @pytest.mark.asyncio
    async def test_no_pysaml2_raises(self):
        from src.core.shared.auth.saml_handler import SAMLHandler
        from src.core.shared.auth.saml_types import SAMLError

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://m")
        with patch("src.core.shared.auth.saml_handler.HAS_PYSAML2", False):
            with pytest.raises(SAMLError, match="PySAML2"):
                await handler._get_saml_client("okta")

    @pytest.mark.asyncio
    async def test_returns_cached_client(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://m")
        mock_client = MagicMock()
        handler._saml_clients["okta"] = mock_client
        with patch("src.core.shared.auth.saml_handler.HAS_PYSAML2", True):
            result = await handler._get_saml_client("okta")
            assert result is mock_client

    @pytest.mark.asyncio
    async def test_creates_new_client(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://m")

        mock_config = MagicMock()
        mock_client = MagicMock()

        with (
            patch("src.core.shared.auth.saml_handler.HAS_PYSAML2", True),
            patch.object(handler, "_fetch_metadata", new_callable=AsyncMock, return_value="<xml/>"),
            patch("src.core.shared.auth.saml_handler.Saml2Config", return_value=mock_config),
            patch("src.core.shared.auth.saml_handler.Saml2Client", return_value=mock_client),
        ):
            result = await handler._get_saml_client("okta")
            assert result is mock_client
            assert handler._saml_clients["okta"] is mock_client


class TestSAMLHandlerInitiateLogin:
    """Cover initiate_login lines 576-632."""

    @pytest.mark.asyncio
    async def test_initiate_login_success(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://m", entity_id="urn:idp")

        mock_client = MagicMock()
        mock_client.prepare_for_authenticate.return_value = (
            "req_123",
            {"headers": [("Location", "http://idp/login?SAMLRequest=abc")]},
        )

        with (
            patch.object(
                handler, "_get_saml_client", new_callable=AsyncMock, return_value=mock_client
            ),
        ):
            url, req_id = await handler.initiate_login("okta", relay_state="/dashboard")
            assert url == "http://idp/login?SAMLRequest=abc"
            assert req_id == "req_123"

    @pytest.mark.asyncio
    async def test_initiate_login_no_entity_id_from_metadata(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://m")

        mock_client = MagicMock()
        mock_client.metadata.identity_providers.return_value = ["urn:idp:auto"]
        mock_client.prepare_for_authenticate.return_value = (
            "req_456",
            {"headers": [("Location", "http://idp/login")]},
        )

        with patch.object(
            handler, "_get_saml_client", new_callable=AsyncMock, return_value=mock_client
        ):
            url, _req_id = await handler.initiate_login("okta")
            assert url == "http://idp/login"

    @pytest.mark.asyncio
    async def test_initiate_login_no_entity_id_raises(self):
        from src.core.shared.auth.saml_config import SAMLConfigurationError
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://m")

        mock_client = MagicMock()
        mock_client.metadata.identity_providers.return_value = []

        with (
            patch.object(
                handler, "_get_saml_client", new_callable=AsyncMock, return_value=mock_client
            ),
            pytest.raises(SAMLConfigurationError, match="Cannot determine entity ID"),
        ):
            await handler.initiate_login("okta")

    @pytest.mark.asyncio
    async def test_initiate_login_no_redirect_url(self):
        from src.core.shared.auth.saml_handler import SAMLHandler
        from src.core.shared.auth.saml_types import SAMLError

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://m", entity_id="urn:idp")

        mock_client = MagicMock()
        mock_client.prepare_for_authenticate.return_value = (
            "req_789",
            {"headers": [("X-Other", "value")]},
        )

        class _NarrowExc(Exception):
            pass

        with (
            patch.object(
                handler, "_get_saml_client", new_callable=AsyncMock, return_value=mock_client
            ),
            patch("src.core.shared.auth.saml_handler.UnknownPrincipal", _NarrowExc),
            patch("src.core.shared.auth.saml_handler.UnsupportedBinding", _NarrowExc),
            pytest.raises(SAMLError, match="redirect URL"),
        ):
            await handler.initiate_login("okta")

    @pytest.mark.asyncio
    async def test_initiate_login_unknown_principal(self):
        from src.core.shared.auth.saml_config import SAMLConfigurationError
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://m", entity_id="urn:idp")

        mock_client = MagicMock()
        # Use the fallback exception class
        from src.core.shared.auth.saml_handler import UnknownPrincipal

        mock_client.prepare_for_authenticate.side_effect = UnknownPrincipal("bad principal")

        with (
            patch.object(
                handler, "_get_saml_client", new_callable=AsyncMock, return_value=mock_client
            ),
            pytest.raises(SAMLConfigurationError, match="configuration error"),
        ):
            await handler.initiate_login("okta")

    @pytest.mark.asyncio
    async def test_initiate_login_runtime_error(self):
        """Cover the generic error handler (line 631-632).

        Note: When pysaml2 is not installed, UnknownPrincipal = Exception,
        so RuntimeError is caught by the first handler. Use a patched
        UnknownPrincipal that doesn't catch our test error.
        """
        from src.core.shared.auth.saml_handler import SAMLHandler
        from src.core.shared.auth.saml_types import SAMLError

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://m", entity_id="urn:idp")

        mock_client = MagicMock()
        mock_client.prepare_for_authenticate.side_effect = TypeError("type fail")

        # Patch UnknownPrincipal/UnsupportedBinding to narrow exception classes
        # so our TypeError falls through to the generic handler
        class _NarrowExc(Exception):
            pass

        with (
            patch.object(
                handler, "_get_saml_client", new_callable=AsyncMock, return_value=mock_client
            ),
            patch("src.core.shared.auth.saml_handler.UnknownPrincipal", _NarrowExc),
            patch("src.core.shared.auth.saml_handler.UnsupportedBinding", _NarrowExc),
            pytest.raises(SAMLError, match="Failed to initiate"),
        ):
            await handler.initiate_login("okta")

    @pytest.mark.asyncio
    async def test_initiate_login_idp_entity_detect_oserror(self):
        """Cover the OSError/PermissionError/RuntimeError catch in entity_id detection."""
        from src.core.shared.auth.saml_config import SAMLConfigurationError
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://m")

        mock_client = MagicMock()
        mock_client.metadata.identity_providers.side_effect = OSError("disk error")

        with (
            patch.object(
                handler, "_get_saml_client", new_callable=AsyncMock, return_value=mock_client
            ),
            pytest.raises(SAMLConfigurationError, match="Cannot determine"),
        ):
            await handler.initiate_login("okta")


class TestSAMLHandlerProcessACS:
    """Cover process_acs_response lines 645-692."""

    @pytest.mark.asyncio
    async def test_process_no_pysaml2(self):
        from src.core.shared.auth.saml_handler import SAMLHandler
        from src.core.shared.auth.saml_types import SAMLError

        handler = SAMLHandler()
        with patch("src.core.shared.auth.saml_handler.HAS_PYSAML2", False):
            with pytest.raises(SAMLError, match="PySAML2"):
                await handler.process_acs_response("response")

    @pytest.mark.asyncio
    async def test_process_success(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://m")
        req_id = handler.store_outstanding_request(idp_name="okta")

        mock_name_id = MagicMock()
        mock_name_id.__str__ = lambda s: "user@test.com"
        mock_name_id.format = None

        mock_response = MagicMock()
        mock_response.name_id = mock_name_id
        mock_response.session_info.return_value = {}
        mock_response.ava = {"email": ["user@test.com"]}
        mock_response.issuer.return_value = "urn:idp"

        mock_client = MagicMock()
        mock_client.parse_authn_request_response.return_value = mock_response

        with (
            patch("src.core.shared.auth.saml_handler.HAS_PYSAML2", True),
            patch.object(
                handler, "_get_saml_client", new_callable=AsyncMock, return_value=mock_client
            ),
        ):
            info = await handler.process_acs_response(
                "response", request_id=req_id, idp_name="okta"
            )
            assert info.email == "user@test.com"

    @pytest.mark.asyncio
    async def test_process_none_response(self):
        from src.core.shared.auth.saml_handler import SAMLHandler
        from src.core.shared.auth.saml_types import SAMLValidationError

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://m")

        mock_client = MagicMock()
        mock_client.parse_authn_request_response.return_value = None

        with (
            patch("src.core.shared.auth.saml_handler.HAS_PYSAML2", True),
            patch.object(
                handler, "_get_saml_client", new_callable=AsyncMock, return_value=mock_client
            ),
            pytest.raises(SAMLValidationError, match="Failed to parse"),
        ):
            await handler.process_acs_response("response", idp_name="okta")

    @pytest.mark.asyncio
    async def test_process_replay_detected(self):
        from src.core.shared.auth.saml_handler import SAMLHandler
        from src.core.shared.auth.saml_types import SAMLReplayError

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://m")

        mock_name_id = MagicMock()
        mock_name_id.__str__ = lambda s: "u"
        mock_name_id.format = None

        mock_response = MagicMock()
        mock_response.name_id = mock_name_id
        mock_response.session_info.return_value = {}
        mock_response.ava = {}
        mock_response.issuer.return_value = None

        mock_client = MagicMock()
        mock_client.parse_authn_request_response.return_value = mock_response

        with (
            patch("src.core.shared.auth.saml_handler.HAS_PYSAML2", True),
            patch.object(
                handler, "_get_saml_client", new_callable=AsyncMock, return_value=mock_client
            ),
            pytest.raises(SAMLReplayError),
        ):
            # Pass a request_id that doesn't exist -> replay
            await handler.process_acs_response("r", request_id="fake_id", idp_name="okta")

    @pytest.mark.asyncio
    async def test_process_runtime_error(self):
        from src.core.shared.auth.saml_handler import SAMLHandler
        from src.core.shared.auth.saml_types import SAMLValidationError

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://m")

        mock_client = MagicMock()
        mock_client.parse_authn_request_response.side_effect = RuntimeError("bad")

        with (
            patch("src.core.shared.auth.saml_handler.HAS_PYSAML2", True),
            patch.object(
                handler, "_get_saml_client", new_callable=AsyncMock, return_value=mock_client
            ),
            pytest.raises(SAMLValidationError, match="validation failed"),
        ):
            await handler.process_acs_response("r", idp_name="okta")

    @pytest.mark.asyncio
    async def test_detect_idp_from_request(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://m")
        req_id = handler.store_outstanding_request(idp_name="okta")

        result = handler._detect_idp_name(req_id)
        assert result == "okta"

    def test_detect_idp_no_request_uses_first(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        handler.register_idp(name="first", metadata_url="http://m")
        handler.register_idp(name="second", metadata_url="http://m2")

        result = handler._detect_idp_name(None)
        assert result == "first"

    def test_detect_idp_no_idps_raises(self):
        from src.core.shared.auth.saml_config import SAMLConfigurationError
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        with pytest.raises(SAMLConfigurationError, match="No IdPs"):
            handler._detect_idp_name(None)


class TestSAMLHandlerInitiateLogout:
    """Cover initiate_logout lines 750-798."""

    @pytest.mark.asyncio
    async def test_no_pysaml2_returns_none(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://m")
        with patch("src.core.shared.auth.saml_handler.HAS_PYSAML2", False):
            result = await handler.initiate_logout("okta", name_id="user")
            assert result is None

    @pytest.mark.asyncio
    async def test_no_slo_url_returns_none(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://m")
        with patch("src.core.shared.auth.saml_handler.HAS_PYSAML2", True):
            result = await handler.initiate_logout("okta", name_id="user")
            assert result is None

    @pytest.mark.asyncio
    async def test_logout_success(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        handler.register_idp(
            name="okta", metadata_url="http://m", entity_id="urn:idp", slo_url="http://slo"
        )

        mock_client = MagicMock()
        mock_client.do_logout.return_value = (
            "logout_req",
            {"headers": [("Location", "http://idp/logout")]},
        )

        with (
            patch("src.core.shared.auth.saml_handler.HAS_PYSAML2", True),
            patch.object(
                handler, "_get_saml_client", new_callable=AsyncMock, return_value=mock_client
            ),
        ):
            result = await handler.initiate_logout(
                "okta", name_id="user", session_index="_sess", relay_state="/home"
            )
            assert "http://idp/logout" in result
            assert "RelayState" in result

    @pytest.mark.asyncio
    async def test_logout_no_entity_id_from_metadata(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://m", slo_url="http://slo")

        mock_client = MagicMock()
        mock_client.metadata.identity_providers.return_value = []

        with (
            patch("src.core.shared.auth.saml_handler.HAS_PYSAML2", True),
            patch.object(
                handler, "_get_saml_client", new_callable=AsyncMock, return_value=mock_client
            ),
        ):
            result = await handler.initiate_logout("okta", name_id="user")
            assert result is None

    @pytest.mark.asyncio
    async def test_logout_with_entity_from_metadata(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://m", slo_url="http://slo")

        mock_client = MagicMock()
        mock_client.metadata.identity_providers.return_value = ["urn:auto"]
        mock_client.do_logout.return_value = (
            "req",
            {"headers": [("Location", "http://idp/logout?q=1")]},
        )

        with (
            patch("src.core.shared.auth.saml_handler.HAS_PYSAML2", True),
            patch.object(
                handler, "_get_saml_client", new_callable=AsyncMock, return_value=mock_client
            ),
        ):
            result = await handler.initiate_logout("okta", name_id="user", relay_state="/dashboard")
            assert "RelayState" in result
            assert "&" in result  # Uses & separator since ? already exists

    @pytest.mark.asyncio
    async def test_logout_error_returns_none(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        handler.register_idp(
            name="okta", metadata_url="http://m", entity_id="urn:idp", slo_url="http://slo"
        )

        mock_client = MagicMock()
        mock_client.do_logout.side_effect = RuntimeError("fail")

        with (
            patch("src.core.shared.auth.saml_handler.HAS_PYSAML2", True),
            patch.object(
                handler, "_get_saml_client", new_callable=AsyncMock, return_value=mock_client
            ),
        ):
            result = await handler.initiate_logout("okta", name_id="user")
            assert result is None

    @pytest.mark.asyncio
    async def test_logout_no_location_no_relay(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        handler.register_idp(
            name="okta", metadata_url="http://m", entity_id="urn:idp", slo_url="http://slo"
        )

        mock_client = MagicMock()
        mock_client.do_logout.return_value = ("req", {"headers": []})

        with (
            patch("src.core.shared.auth.saml_handler.HAS_PYSAML2", True),
            patch.object(
                handler, "_get_saml_client", new_callable=AsyncMock, return_value=mock_client
            ),
        ):
            result = await handler.initiate_logout("okta", name_id="user")
            assert result is None


class TestSAMLHandlerProcessSLS:
    """Cover process_sls_response lines 817-848."""

    @pytest.mark.asyncio
    async def test_no_pysaml2_returns_true(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        with patch("src.core.shared.auth.saml_handler.HAS_PYSAML2", False):
            result = await handler.process_sls_response("resp", "okta")
            assert result is True

    @pytest.mark.asyncio
    async def test_sls_success(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://m")

        mock_response = MagicMock()
        mock_response.status_ok.return_value = True

        mock_client = MagicMock()
        mock_client.parse_logout_request_response.return_value = mock_response

        with (
            patch("src.core.shared.auth.saml_handler.HAS_PYSAML2", True),
            patch.object(
                handler, "_get_saml_client", new_callable=AsyncMock, return_value=mock_client
            ),
        ):
            result = await handler.process_sls_response("resp", "okta")
            assert result is True

    @pytest.mark.asyncio
    async def test_sls_error_returns_false(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://m")

        mock_client = MagicMock()
        mock_client.parse_logout_request_response.side_effect = RuntimeError("fail")

        with (
            patch("src.core.shared.auth.saml_handler.HAS_PYSAML2", True),
            patch.object(
                handler, "_get_saml_client", new_callable=AsyncMock, return_value=mock_client
            ),
        ):
            result = await handler.process_sls_response("resp", "okta")
            assert result is False


class TestSAMLHandlerGenerateMetadata:
    """Cover generate_metadata lines 904-929."""

    @pytest.mark.asyncio
    async def test_generate_metadata_no_pysaml2(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        with patch("src.core.shared.auth.saml_handler.HAS_PYSAML2", False):
            metadata = await handler.generate_metadata()
            assert "EntityDescriptor" in metadata
            assert "urn:acgs2:saml:sp" in metadata

    @pytest.mark.asyncio
    async def test_generate_metadata_with_pysaml2_and_idps(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        handler.register_idp(name="okta", metadata_url="http://m")

        mock_client = MagicMock()

        # create_metadata_string may not exist in module scope if pysaml2 not installed
        with (
            patch("src.core.shared.auth.saml_handler.HAS_PYSAML2", True),
            patch.object(
                handler, "_get_saml_client", new_callable=AsyncMock, return_value=mock_client
            ),
            patch(
                "src.core.shared.auth.saml_handler.create_metadata_string",
                create=True,
                return_value=b"<metadata/>",
            ),
        ):
            result = await handler.generate_metadata()
            assert result == "<metadata/>"

    @pytest.mark.asyncio
    async def test_generate_metadata_with_pysaml2_no_idps(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()

        mock_config = MagicMock()
        mock_client = MagicMock()

        with (
            patch("src.core.shared.auth.saml_handler.HAS_PYSAML2", True),
            patch("src.core.shared.auth.saml_handler.Saml2Config", return_value=mock_config),
            patch("src.core.shared.auth.saml_handler.Saml2Client", return_value=mock_client),
            patch(
                "src.core.shared.auth.saml_handler.create_metadata_string",
                create=True,
                return_value="<str_metadata/>",
            ),
        ):
            result = await handler.generate_metadata()
            assert result == "<str_metadata/>"

    @pytest.mark.asyncio
    async def test_generate_metadata_error(self):
        from src.core.shared.auth.saml_handler import SAMLHandler
        from src.core.shared.auth.saml_types import SAMLError

        handler = SAMLHandler()

        with (
            patch("src.core.shared.auth.saml_handler.HAS_PYSAML2", True),
            patch(
                "src.core.shared.auth.saml_handler.Saml2Config",
                side_effect=RuntimeError("config fail"),
            ),
            pytest.raises(SAMLError, match="Failed to generate"),
        ):
            await handler.generate_metadata()


class TestSAMLHandlerClose:
    """Cover close() cleanup."""

    @pytest.mark.asyncio
    async def test_close_cleans_up(self):
        from src.core.shared.auth.saml_handler import SAMLHandler

        handler = SAMLHandler()
        mock_client = AsyncMock()
        handler._http_client = mock_client

        # Set temp files
        temp_cert = MagicMock()
        temp_cert.name = "/tmp/c.crt"
        handler._temp_cert_file = temp_cert

        with patch("pathlib.Path.unlink"):
            await handler.close()
            mock_client.aclose.assert_awaited_once()
            assert handler._http_client is None


# ---------------------------------------------------------------------------
# provisioning.py tests
# ---------------------------------------------------------------------------


class TestJITProvisionerInMemory:
    """Cover _provision_in_memory and get_or_create_user without session."""

    @pytest.mark.asyncio
    async def test_in_memory_provision(self):
        from src.core.shared.auth.provisioning import JITProvisioner

        p = JITProvisioner(default_roles=["viewer"])
        result = await p.get_or_create_user(
            email="USER@Example.com",
            name="Test User",
            sso_provider="saml",
            idp_user_id="idp-123",
            provider_id="prov-456",
            roles=["admin"],
            name_id="user@example.com",
            session_index="_sess",
        )
        assert result.created is True
        assert result.user["email"] == "user@example.com"
        assert result.user["sso_provider"] == "saml"
        assert "admin" in result.user["roles"]

    @pytest.mark.asyncio
    async def test_in_memory_default_roles(self):
        from src.core.shared.auth.provisioning import JITProvisioner

        p = JITProvisioner(default_roles=["viewer"])
        result = await p.get_or_create_user(email="a@b.com")
        assert "viewer" in result.user["roles"]

    @pytest.mark.asyncio
    async def test_domain_not_allowed(self):
        from src.core.shared.auth.provisioning import DomainNotAllowedError, JITProvisioner

        p = JITProvisioner(allowed_domains=["allowed.com"])
        with pytest.raises(DomainNotAllowedError, match="not allowed"):
            await p.get_or_create_user(email="user@blocked.com")


def _setup_provisioning_user_module():
    """Mock the src.core.shared.models.user module for provisioning tests."""
    mock_user_module = MagicMock()

    class MockSSOProviderType:
        SAML = MagicMock()
        SAML.value = "saml"
        OIDC = MagicMock()
        OIDC.value = "oidc"

    mock_user_module.SSOProviderType = MockSSOProviderType
    mock_user_module.User = MagicMock()
    return mock_user_module


def _provisioning_orm_patches(mock_user_module):
    """Return context managers for provisioning ORM tests.

    Must be used with `with` statement. Patches:
    - src.core.shared.models.user in sys.modules
    - sqlalchemy.select
    - sqlalchemy.orm.joinedload
    """
    mock_select = MagicMock()
    mock_select.return_value.options.return_value.where.return_value = "stmt"

    return (
        patch.dict(sys.modules, {"src.core.shared.models.user": mock_user_module}),
        patch("sqlalchemy.select", mock_select),
        patch("sqlalchemy.orm.joinedload", MagicMock()),
    )


class TestJITProvisionerORM:
    """Cover _provision_with_orm lines 330-456."""

    @pytest.mark.asyncio
    async def test_provision_new_user_saml(self):
        from src.core.shared.auth.provisioning import JITProvisioner

        mock_user_module = _setup_provisioning_user_module()
        mock_user_instance = MagicMock()
        mock_user_instance.id = "u-1"
        mock_user_instance.email = "test@example.com"
        mock_user_instance.name = "Test"
        mock_user_instance.sso_enabled = True
        mock_user_instance.sso_provider = mock_user_module.SSOProviderType.SAML
        mock_user_instance.sso_idp_user_id = "idp-1"
        mock_user_instance.sso_provider_id = "p-1"
        mock_user_instance.sso_name_id = "nid"
        mock_user_instance.sso_session_index = "si"
        mock_user_instance.role_list = ["admin"]
        mock_user_instance.created_at = datetime(2024, 1, 1, tzinfo=UTC)
        mock_user_instance.last_login = datetime.now(UTC)
        mock_user_module.User.return_value = mock_user_instance

        mock_result = MagicMock()
        mock_result.unique.return_value.scalar_one_or_none.return_value = None

        mock_session = AsyncMock()
        mock_session.execute.return_value = mock_result
        mock_session.add = MagicMock()

        p = JITProvisioner(default_roles=["viewer"])
        p1, p2, p3 = _provisioning_orm_patches(mock_user_module)

        with p1, p2, p3:
            result = await p._provision_with_orm(
                session=mock_session,
                email="test@example.com",
                name="Test",
                sso_provider="saml",
                idp_user_id="idp-1",
                provider_id="p-1",
                roles=["admin"],
                name_id="nid",
                session_index="si",
            )
            assert result.created is True
            assert result.user["email"] == "test@example.com"
            mock_session.add.assert_called_once()

    @pytest.mark.asyncio
    async def test_provision_existing_user_oidc(self):
        from src.core.shared.auth.provisioning import JITProvisioner

        mock_user_module = _setup_provisioning_user_module()

        existing_user = MagicMock()
        existing_user.id = "u-2"
        existing_user.email = "existing@example.com"
        existing_user.name = "Old Name"
        existing_user.sso_enabled = True
        existing_user.sso_provider = mock_user_module.SSOProviderType.OIDC
        existing_user.sso_idp_user_id = "old-id"
        existing_user.sso_provider_id = "p-2"
        existing_user.sso_name_id = None
        existing_user.sso_session_index = None
        existing_user.role_list = ["viewer"]
        existing_user.created_at = datetime(2024, 1, 1, tzinfo=UTC)
        existing_user.last_login = datetime(2024, 6, 1, tzinfo=UTC)

        mock_result = MagicMock()
        mock_result.unique.return_value.scalar_one_or_none.return_value = existing_user

        mock_session = AsyncMock()
        mock_session.execute.return_value = mock_result

        p = JITProvisioner()
        p1, p2, p3 = _provisioning_orm_patches(mock_user_module)

        with p1, p2, p3:
            result = await p._provision_with_orm(
                session=mock_session,
                email="existing@example.com",
                name="New Name",
                sso_provider="oidc",
                idp_user_id="new-id",
                provider_id="p-2",
                roles=["admin", "developer"],
                name_id=None,
                session_index=None,
            )
            assert result.created is False
            assert result.roles_updated is True
            existing_user.update_sso_info.assert_called_once()
            existing_user.set_roles.assert_called_once()

    @pytest.mark.asyncio
    async def test_provision_disabled_raises(self):
        from src.core.shared.auth.provisioning import JITProvisioner, ProvisioningDisabledError

        mock_user_module = _setup_provisioning_user_module()

        mock_result = MagicMock()
        mock_result.unique.return_value.scalar_one_or_none.return_value = None

        mock_session = AsyncMock()
        mock_session.execute.return_value = mock_result

        p = JITProvisioner(auto_provision_enabled=False)
        p1, p2, p3 = _provisioning_orm_patches(mock_user_module)

        with p1, p2, p3, pytest.raises(ProvisioningDisabledError):
            await p._provision_with_orm(
                session=mock_session,
                email="test@example.com",
                name=None,
                sso_provider="oidc",
                idp_user_id=None,
                provider_id=None,
                roles=None,
                name_id=None,
                session_index=None,
            )

    @pytest.mark.asyncio
    async def test_provision_existing_no_role_change(self):
        from src.core.shared.auth.provisioning import JITProvisioner

        mock_user_module = _setup_provisioning_user_module()

        existing_user = MagicMock()
        existing_user.id = "u-3"
        existing_user.email = "same@example.com"
        existing_user.name = "Same"
        existing_user.sso_enabled = True
        existing_user.sso_provider = mock_user_module.SSOProviderType.OIDC
        existing_user.sso_idp_user_id = "id"
        existing_user.sso_provider_id = "p"
        existing_user.sso_name_id = None
        existing_user.sso_session_index = None
        existing_user.role_list = ["viewer"]
        existing_user.created_at = datetime(2024, 1, 1, tzinfo=UTC)
        existing_user.last_login = datetime.now(UTC)

        mock_result = MagicMock()
        mock_result.unique.return_value.scalar_one_or_none.return_value = existing_user

        mock_session = AsyncMock()
        mock_session.execute.return_value = mock_result

        p = JITProvisioner()
        p1, p2, p3 = _provisioning_orm_patches(mock_user_module)

        with p1, p2, p3:
            result = await p._provision_with_orm(
                session=mock_session,
                email="same@example.com",
                name="Same",  # Same name, no change
                sso_provider="oidc",
                idp_user_id=None,
                provider_id=None,
                roles=None,  # No roles from IdP
                name_id=None,
                session_index=None,
            )
            assert result.created is False
            assert result.roles_updated is False


class TestJITProvisionerUpdateRoles:
    """Cover update_user_roles lines 551-582."""

    @pytest.mark.asyncio
    async def test_update_user_roles_not_found(self):
        from src.core.shared.auth.provisioning import JITProvisioner, ProvisioningError

        mock_user_module = _setup_provisioning_user_module()

        mock_result = MagicMock()
        mock_result.unique.return_value.scalar_one_or_none.return_value = None

        mock_session = AsyncMock()
        mock_session.execute.return_value = mock_result

        p = JITProvisioner()
        p1, p2, p3 = _provisioning_orm_patches(mock_user_module)

        with p1, p2, p3, pytest.raises(ProvisioningError, match="not found"):
            await p.update_user_roles("u-1", ["admin"], mock_session)

    @pytest.mark.asyncio
    async def test_update_user_roles_changed(self):
        from src.core.shared.auth.provisioning import JITProvisioner

        mock_user_module = _setup_provisioning_user_module()

        existing_user = MagicMock()
        existing_user.role_list = ["viewer"]

        mock_result = MagicMock()
        mock_result.unique.return_value.scalar_one_or_none.return_value = existing_user

        mock_session = AsyncMock()
        mock_session.execute.return_value = mock_result

        p = JITProvisioner()
        p1, p2, p3 = _provisioning_orm_patches(mock_user_module)

        with p1, p2, p3:
            changed = await p.update_user_roles("u-1", ["admin"], mock_session)
            assert changed is True
            existing_user.set_roles.assert_called_once()

    @pytest.mark.asyncio
    async def test_update_user_roles_no_change(self):
        from src.core.shared.auth.provisioning import JITProvisioner

        mock_user_module = _setup_provisioning_user_module()

        existing_user = MagicMock()
        existing_user.role_list = ["admin"]

        mock_result = MagicMock()
        mock_result.unique.return_value.scalar_one_or_none.return_value = existing_user

        mock_session = AsyncMock()
        mock_session.execute.return_value = mock_result

        p = JITProvisioner()
        p1, p2, p3 = _provisioning_orm_patches(mock_user_module)

        with p1, p2, p3:
            # Pass same roles - no change
            changed = await p.update_user_roles("u-1", ["admin"], mock_session)
            assert changed is False


class TestJITProvisionerClearSession:
    """Cover clear_sso_session lines 601-615."""

    @pytest.mark.asyncio
    async def test_clear_session_not_found(self):
        from src.core.shared.auth.provisioning import JITProvisioner, ProvisioningError

        mock_user_module = _setup_provisioning_user_module()

        mock_result = MagicMock()
        mock_result.unique.return_value.scalar_one_or_none.return_value = None

        mock_session = AsyncMock()
        mock_session.execute.return_value = mock_result

        p = JITProvisioner()
        p1, p2, p3 = _provisioning_orm_patches(mock_user_module)

        with p1, p2, p3, pytest.raises(ProvisioningError, match="not found"):
            await p.clear_sso_session("u-1", mock_session)

    @pytest.mark.asyncio
    async def test_clear_session_success(self):
        from src.core.shared.auth.provisioning import JITProvisioner

        mock_user_module = _setup_provisioning_user_module()

        existing_user = MagicMock()

        mock_result = MagicMock()
        mock_result.unique.return_value.scalar_one_or_none.return_value = existing_user

        mock_session = AsyncMock()
        mock_session.execute.return_value = mock_result

        p = JITProvisioner()
        p1, p2, p3 = _provisioning_orm_patches(mock_user_module)

        with p1, p2, p3:
            await p.clear_sso_session("u-1", mock_session)
            existing_user.clear_sso_session.assert_called_once()


class TestProvisionerSingleton:
    """Cover get_provisioner / reset_provisioner."""

    def test_singleton(self):
        from src.core.shared.auth.provisioning import (
            get_provisioner,
            reset_provisioner,
        )

        reset_provisioner()
        p1 = get_provisioner(default_roles=["admin"])
        p2 = get_provisioner()
        assert p1 is p2
        reset_provisioner()


class TestMergeRoles:
    """Cover _merge_roles edge cases."""

    def test_no_new_roles_no_existing_with_defaults(self):
        from src.core.shared.auth.provisioning import JITProvisioner

        p = JITProvisioner(default_roles=["viewer"])
        merged, changed = p._merge_roles([], [], ["viewer"])
        assert merged == ["viewer"]
        assert changed is True

    def test_existing_preserved_when_no_new(self):
        from src.core.shared.auth.provisioning import JITProvisioner

        p = JITProvisioner()
        merged, changed = p._merge_roles(["admin"], [])
        assert merged == ["admin"]
        assert changed is False

    def test_new_roles_override(self):
        from src.core.shared.auth.provisioning import JITProvisioner

        p = JITProvisioner()
        merged, changed = p._merge_roles(["old"], ["new"])
        assert merged == ["new"]
        assert changed is True


# ---------------------------------------------------------------------------
# role_mapper.py tests
# ---------------------------------------------------------------------------


class TestRoleMapperCaseSensitive:
    """Cover case_sensitive path line 194."""

    def test_case_sensitive_match(self):
        from src.core.shared.auth.role_mapper import RoleMapper

        mapper = RoleMapper(default_mappings={"Admin": "admin"}, case_sensitive=True)
        # Exact case match
        assert mapper.map_groups(["Admin"]) == ["admin"]
        # Wrong case does not match
        assert mapper.map_groups(["admin"]) == []

    def test_case_sensitive_normalize(self):
        from src.core.shared.auth.role_mapper import RoleMapper

        mapper = RoleMapper(case_sensitive=True)
        assert mapper._normalize_group("  Test  ") == "Test"

    def test_case_insensitive_normalize(self):
        from src.core.shared.auth.role_mapper import RoleMapper

        mapper = RoleMapper(case_sensitive=False)
        assert mapper._normalize_group("  TEST  ") == "test"


class TestRoleMapperAsyncMapping:
    """Cover map_groups_async lines 313-318, 333-344, 351, 355."""

    @pytest.mark.asyncio
    async def test_empty_groups(self):
        from src.core.shared.auth.role_mapper import RoleMapper

        mapper = RoleMapper()
        result = await mapper.map_groups_async([])
        assert result.roles == []
        assert result.source == "none"

    @pytest.mark.asyncio
    async def test_default_mapping_only(self):
        from src.core.shared.auth.role_mapper import RoleMapper

        mapper = RoleMapper()
        result = await mapper.map_groups_async(["admins", "unknown_group"])
        assert "admin" in result.roles
        assert "unknown_group" in result.unmapped_groups
        assert result.source == "default"

    @pytest.mark.asyncio
    async def test_db_mapping_success(self):
        from src.core.shared.auth.role_mapper import RoleMapper

        mapper = RoleMapper()
        mock_session = AsyncMock()

        with patch.object(
            mapper,
            "_fetch_provider_mappings",
            new_callable=AsyncMock,
            return_value={"engineering": ("developer", 10)},
        ):
            result = await mapper.map_groups_async(
                ["engineering", "unknown"],
                provider_id="p-1",
                session=mock_session,
            )
            assert "developer" in result.roles
            assert result.source == "database"

    @pytest.mark.asyncio
    async def test_db_mapping_fallback_on_error(self):
        from src.core.shared.auth.role_mapper import RoleMapper

        mapper = RoleMapper()
        mock_session = AsyncMock()

        with patch.object(
            mapper,
            "_fetch_provider_mappings",
            new_callable=AsyncMock,
            side_effect=RuntimeError("db error"),
        ):
            result = await mapper.map_groups_async(
                ["admins"],
                provider_id="p-1",
                session=mock_session,
            )
            assert "admin" in result.roles
            assert result.source == "default"

    @pytest.mark.asyncio
    async def test_db_mapping_with_fallback_role(self):
        from src.core.shared.auth.role_mapper import RoleMapper

        mapper = RoleMapper(fallback_role="viewer")
        mock_session = AsyncMock()

        with patch.object(
            mapper,
            "_fetch_provider_mappings",
            new_callable=AsyncMock,
            return_value={"eng": ("developer", 10)},
        ):
            result = await mapper.map_groups_async(
                ["unknown"],
                provider_id="p-1",
                session=mock_session,
            )
            assert "viewer" in result.roles

    @pytest.mark.asyncio
    async def test_db_mapping_case_sensitive(self):
        from src.core.shared.auth.role_mapper import RoleMapper

        mapper = RoleMapper(case_sensitive=True)
        mock_session = AsyncMock()

        with patch.object(
            mapper,
            "_fetch_provider_mappings",
            new_callable=AsyncMock,
            return_value={"Engineering": ("developer", 10)},
        ):
            result = await mapper.map_groups_async(
                ["Engineering"],
                provider_id="p-1",
                session=mock_session,
            )
            assert "developer" in result.roles


def _setup_role_mapper_db_mocks():
    """Set up sys.modules mocks for role_mapper DB tests.

    Returns (mock_module, mock_select) context managers are NOT used --
    caller must use patch.dict on sys.modules.
    """
    mock_module = MagicMock()
    mock_module.SSORoleMapping = MagicMock()
    mock_module.SSORoleMapping.provider_id = MagicMock()
    mock_module.SSORoleMapping.priority = MagicMock()
    mock_module.SSORoleMapping.idp_group = MagicMock()
    mock_module.SSORoleMapping.id = MagicMock()
    mock_module.SSORoleMapping.create_mapping = MagicMock()
    return mock_module


class TestRoleMapperFetchProviderMappings:
    """Cover _fetch_provider_mappings lines 392-405."""

    @pytest.mark.asyncio
    async def test_fetch_provider_mappings(self):
        from src.core.shared.auth.role_mapper import RoleMapper

        mapper = RoleMapper()

        mock_mapping = MagicMock()
        mock_mapping.idp_group = "eng"
        mock_mapping.acgs_role = "developer"
        mock_mapping.priority = 5

        mock_result = MagicMock()
        mock_result.scalars.return_value.all.return_value = [mock_mapping]

        mock_session = AsyncMock()
        mock_session.execute.return_value = mock_result

        mock_module = _setup_role_mapper_db_mocks()
        mock_select = MagicMock()
        mock_select.return_value.where.return_value.order_by.return_value = "stmt"

        with (
            patch.dict(sys.modules, {"src.core.shared.models.sso_role_mapping": mock_module}),
            patch("sqlalchemy.select", mock_select),
        ):
            result = await mapper._fetch_provider_mappings("p-1", mock_session)
            assert result == {"eng": ("developer", 5)}


class TestRoleMapperCreateMapping:
    """Cover create_mapping lines 478-502."""

    @pytest.mark.asyncio
    async def test_create_mapping(self):
        from src.core.shared.auth.role_mapper import RoleMapper

        mapper = RoleMapper()
        mock_session = MagicMock()

        mock_mapping = MagicMock()
        mock_mapping.id = "m-1"

        mock_module = _setup_role_mapper_db_mocks()
        mock_module.SSORoleMapping.create_mapping.return_value = mock_mapping

        with patch.dict(sys.modules, {"src.core.shared.models.sso_role_mapping": mock_module}):
            result = await mapper.create_mapping(
                provider_id="p-1",
                idp_group="eng",
                acgs_role="developer",
                session=mock_session,
                priority=5,
                description="Test",
            )
            assert result is mock_mapping
            mock_session.add.assert_called_once_with(mock_mapping)


class TestRoleMapperDeleteMapping:
    """Cover delete_mapping lines 518-539."""

    @pytest.mark.asyncio
    async def test_delete_mapping_found(self):
        from src.core.shared.auth.role_mapper import RoleMapper

        mapper = RoleMapper()
        mock_mapping = MagicMock()

        mock_result = MagicMock()
        mock_result.scalar_one_or_none.return_value = mock_mapping

        mock_session = AsyncMock()
        mock_session.execute.return_value = mock_result

        mock_module = _setup_role_mapper_db_mocks()
        mock_select = MagicMock()
        mock_select.return_value.where.return_value = "stmt"

        with (
            patch.dict(sys.modules, {"src.core.shared.models.sso_role_mapping": mock_module}),
            patch("sqlalchemy.select", mock_select),
        ):
            result = await mapper.delete_mapping("m-1", mock_session)
            assert result is True
            mock_session.delete.assert_awaited_once_with(mock_mapping)

    @pytest.mark.asyncio
    async def test_delete_mapping_not_found(self):
        from src.core.shared.auth.role_mapper import RoleMapper

        mapper = RoleMapper()
        mock_result = MagicMock()
        mock_result.scalar_one_or_none.return_value = None

        mock_session = AsyncMock()
        mock_session.execute.return_value = mock_result

        mock_module = _setup_role_mapper_db_mocks()
        mock_select = MagicMock()
        mock_select.return_value.where.return_value = "stmt"

        with (
            patch.dict(sys.modules, {"src.core.shared.models.sso_role_mapping": mock_module}),
            patch("sqlalchemy.select", mock_select),
        ):
            result = await mapper.delete_mapping("m-999", mock_session)
            assert result is False


class TestRoleMapperGetProviderMappings:
    """Cover get_provider_mappings lines 555-568."""

    @pytest.mark.asyncio
    async def test_get_provider_mappings(self):
        from src.core.shared.auth.role_mapper import RoleMapper

        mapper = RoleMapper()

        mock_mapping = MagicMock()
        mock_mapping.id = "m-1"
        mock_mapping.provider_id = "p-1"
        mock_mapping.idp_group = "eng"
        mock_mapping.acgs_role = "developer"
        mock_mapping.priority = 5
        mock_mapping.description = "Test"
        mock_mapping.created_at = datetime(2024, 1, 1, tzinfo=UTC)
        mock_mapping.updated_at = None

        mock_result = MagicMock()
        mock_result.scalars.return_value.all.return_value = [mock_mapping]

        mock_session = AsyncMock()
        mock_session.execute.return_value = mock_result

        mock_module = _setup_role_mapper_db_mocks()
        mock_select = MagicMock()
        mock_select.return_value.where.return_value.order_by.return_value = "stmt"

        with (
            patch.dict(sys.modules, {"src.core.shared.models.sso_role_mapping": mock_module}),
            patch("sqlalchemy.select", mock_select),
        ):
            result = await mapper.get_provider_mappings("p-1", mock_session)
            assert len(result) == 1
            assert result[0]["idp_group"] == "eng"
            assert result[0]["acgs_role"] == "developer"
            assert result[0]["created_at"] == "2024-01-01T00:00:00+00:00"
            assert result[0]["updated_at"] is None


class TestRoleMapperAddRemove:
    """Cover add/remove default mapping methods."""

    def test_add_default_mapping(self):
        from src.core.shared.auth.role_mapper import RoleMapper

        mapper = RoleMapper(default_mappings={})
        mapper.add_default_mapping("Custom Group", "custom_role")
        assert mapper.default_mappings["custom group"] == "custom_role"

    def test_add_default_mapping_case_sensitive(self):
        from src.core.shared.auth.role_mapper import RoleMapper

        mapper = RoleMapper(default_mappings={}, case_sensitive=True)
        mapper.add_default_mapping("Custom", "custom_role")
        assert mapper.default_mappings["Custom"] == "custom_role"

    def test_remove_default_mapping(self):
        from src.core.shared.auth.role_mapper import RoleMapper

        mapper = RoleMapper(default_mappings={"test": "role"})
        assert mapper.remove_default_mapping("test") is True
        assert mapper.remove_default_mapping("nonexistent") is False

    def test_get_default_mappings_returns_copy(self):
        from src.core.shared.auth.role_mapper import RoleMapper

        mapper = RoleMapper(default_mappings={"a": "b"})
        mappings = mapper.get_default_mappings()
        assert mappings == {"a": "b"}
        mappings["c"] = "d"
        assert "c" not in mapper.default_mappings


class TestRoleMapperSingleton:
    """Cover get_role_mapper / reset_role_mapper."""

    def test_singleton(self):
        from src.core.shared.auth.role_mapper import get_role_mapper, reset_role_mapper

        reset_role_mapper()
        m1 = get_role_mapper(fallback_role="viewer")
        m2 = get_role_mapper()
        assert m1 is m2
        reset_role_mapper()


class TestRoleMapperFallbackRole:
    """Cover fallback role application in sync map_groups."""

    def test_fallback_applied(self):
        from src.core.shared.auth.role_mapper import RoleMapper

        mapper = RoleMapper(
            default_mappings={},
            fallback_role="viewer",
        )
        roles = mapper.map_groups(["unknown_group"])
        assert roles == ["viewer"]

    def test_no_fallback(self):
        from src.core.shared.auth.role_mapper import RoleMapper

        mapper = RoleMapper(default_mappings={}, fallback_role=None)
        roles = mapper.map_groups(["unknown"])
        assert roles == []

    def test_empty_groups(self):
        from src.core.shared.auth.role_mapper import RoleMapper

        mapper = RoleMapper()
        roles = mapper.map_groups([])
        assert roles == []



# Constitutional Hash: 608508a9bd224290
"""Agent code execution sandbox for OWASP AA07 mitigation.

Provides process-level isolation for agent-generated code execution with:
- Resource limits (CPU time, memory, file descriptors)
- Filesystem restriction (temporary directory only)
- Network isolation (optional)
- Output capture and size limits

Architecture:
- SandboxConfig: frozen configuration for sandbox instances
- SandboxResult: frozen execution result with captured output
- Sandbox: context manager for isolated code execution
- SandboxPolicy: predefined policies (STRICT, STANDARD, PERMISSIVE)

The default backend uses subprocess with resource limits (rlimit).
Can be upgraded to gVisor, E2B, or Firecracker by implementing
the SandboxBackend protocol.
"""

from __future__ import annotations

import resource
import signal
import subprocess
import tempfile
from dataclasses import dataclass, field
from datetime import UTC, datetime
from enum import StrEnum
from pathlib import Path
from typing import Protocol, runtime_checkable

from src.core.shared.structured_logging import get_logger

logger = get_logger(__name__)

# Maximum output size to capture (prevent memory exhaustion)
MAX_OUTPUT_BYTES = 1_048_576  # 1 MB


class SandboxPolicy(StrEnum):
    """Predefined sandbox policies."""

    STRICT = "strict"  # No network, minimal resources, 5s timeout
    STANDARD = "standard"  # No network, moderate resources, 30s timeout
    PERMISSIVE = "permissive"  # Network allowed, generous resources, 120s timeout


@dataclass(frozen=True)
class SandboxConfig:
    """Immutable sandbox configuration."""

    policy: SandboxPolicy = SandboxPolicy.STANDARD
    timeout_seconds: int = 30
    max_memory_bytes: int = 256 * 1024 * 1024  # 256 MB
    max_cpu_seconds: int = 30
    max_file_descriptors: int = 64
    max_output_bytes: int = MAX_OUTPUT_BYTES
    allow_network: bool = False
    allowed_paths: tuple[str, ...] = ()
    env_vars: dict[str, str] = field(default_factory=dict)

    @classmethod
    def from_policy(cls, policy: SandboxPolicy) -> SandboxConfig:
        """Create config from a predefined policy."""
        if policy == SandboxPolicy.STRICT:
            return cls(
                policy=policy,
                timeout_seconds=5,
                max_memory_bytes=64 * 1024 * 1024,
                max_cpu_seconds=5,
                max_file_descriptors=16,
                allow_network=False,
            )
        if policy == SandboxPolicy.PERMISSIVE:
            return cls(
                policy=policy,
                timeout_seconds=120,
                max_memory_bytes=1024 * 1024 * 1024,
                max_cpu_seconds=120,
                max_file_descriptors=256,
                allow_network=True,
            )
        # STANDARD is the default
        return cls(policy=policy)


@dataclass(frozen=True)
class SandboxResult:
    """Immutable result of sandboxed execution."""

    success: bool
    exit_code: int
    stdout: str
    stderr: str
    duration_seconds: float
    timed_out: bool
    memory_exceeded: bool
    sandbox_id: str
    executed_at: datetime = field(default_factory=lambda: datetime.now(UTC))
    error: str | None = None


@runtime_checkable
class SandboxBackend(Protocol):
    """Protocol for sandbox backend implementations.

    Default: ProcessSandbox (subprocess + rlimit)
    Future: gVisorSandbox, E2BSandbox, FirecrackerSandbox
    """

    def execute(
        self,
        code: str,
        config: SandboxConfig,
        sandbox_dir: Path,
    ) -> SandboxResult: ...


def _set_resource_limits(config: SandboxConfig) -> None:
    """Set resource limits for the child process (called via preexec_fn)."""
    # CPU time limit
    resource.setrlimit(
        resource.RLIMIT_CPU,
        (config.max_cpu_seconds, config.max_cpu_seconds + 1),
    )
    # Memory limit (address space)
    resource.setrlimit(
        resource.RLIMIT_AS,
        (config.max_memory_bytes, config.max_memory_bytes),
    )
    # File descriptor limit
    resource.setrlimit(
        resource.RLIMIT_NOFILE,
        (config.max_file_descriptors, config.max_file_descriptors),
    )
    # No core dumps
    resource.setrlimit(resource.RLIMIT_CORE, (0, 0))


class ProcessSandbox:
    """Default sandbox backend using subprocess with resource limits."""

    def execute(
        self,
        code: str,
        config: SandboxConfig,
        sandbox_dir: Path,
    ) -> SandboxResult:
        """Execute code in an isolated subprocess with resource limits."""
        import time
        import uuid

        sandbox_id = str(uuid.uuid4())[:12]
        start_time = time.monotonic()
        timed_out = False
        memory_exceeded = False

        # Write code to temp file in sandbox directory
        code_file = sandbox_dir / f"sandbox_{sandbox_id}.py"
        code_file.write_text(code, encoding="utf-8")

        # Build environment
        env = {
            "HOME": str(sandbox_dir),
            "TMPDIR": str(sandbox_dir),
            "PATH": "/usr/bin:/bin",
            "PYTHONDONTWRITEBYTECODE": "1",
            "PYTHONHASHSEED": "0",
        }
        env.update(config.env_vars)

        try:
            result = subprocess.run(  # noqa: S603 - fixed local interpreter
                ["/usr/bin/python3", str(code_file)],
                capture_output=True,
                timeout=config.timeout_seconds,
                cwd=str(sandbox_dir),
                env=env,
                preexec_fn=lambda: _set_resource_limits(config),
            )

            duration = time.monotonic() - start_time
            stdout = result.stdout.decode("utf-8", errors="replace")[: config.max_output_bytes]
            stderr = result.stderr.decode("utf-8", errors="replace")[: config.max_output_bytes]

            # Check for resource limit signals
            if result.returncode == -signal.SIGKILL:
                memory_exceeded = True
            elif result.returncode == -signal.SIGXCPU:
                timed_out = True

            return SandboxResult(
                success=result.returncode == 0,
                exit_code=result.returncode,
                stdout=stdout,
                stderr=stderr,
                duration_seconds=duration,
                timed_out=timed_out,
                memory_exceeded=memory_exceeded,
                sandbox_id=sandbox_id,
            )

        except subprocess.TimeoutExpired:
            duration = time.monotonic() - start_time
            return SandboxResult(
                success=False,
                exit_code=-1,
                stdout="",
                stderr="",
                duration_seconds=duration,
                timed_out=True,
                memory_exceeded=False,
                sandbox_id=sandbox_id,
                error=f"Execution timed out after {config.timeout_seconds}s",
            )

        except OSError as e:
            duration = time.monotonic() - start_time
            return SandboxResult(
                success=False,
                exit_code=-1,
                stdout="",
                stderr="",
                duration_seconds=duration,
                timed_out=False,
                memory_exceeded=False,
                sandbox_id=sandbox_id,
                error=f"Sandbox execution failed: {e}",
            )

        finally:
            # Clean up code file
            if code_file.exists():
                code_file.unlink()


class Sandbox:
    """High-level sandbox interface for agent code execution.

    Usage:
        config = SandboxConfig.from_policy(SandboxPolicy.STRICT)
        sandbox = Sandbox(config)
        result = sandbox.run("print('hello')")
        assert result.success
        assert result.stdout.strip() == "hello"
    """

    def __init__(
        self,
        config: SandboxConfig | None = None,
        backend: SandboxBackend | None = None,
    ) -> None:
        self._config = config or SandboxConfig()
        self._backend = backend or ProcessSandbox()

    @property
    def config(self) -> SandboxConfig:
        return self._config

    def run(self, code: str) -> SandboxResult:
        """Execute code in the sandbox and return the result.

        Args:
            code: Python code to execute.

        Returns:
            SandboxResult with captured output and metadata.
        """
        if not code.strip():
            return SandboxResult(
                success=True,
                exit_code=0,
                stdout="",
                stderr="",
                duration_seconds=0.0,
                timed_out=False,
                memory_exceeded=False,
                sandbox_id="empty",
            )

        with tempfile.TemporaryDirectory(prefix="acgs2_sandbox_") as tmpdir:
            sandbox_dir = Path(tmpdir)
            result = self._backend.execute(code, self._config, sandbox_dir)

            if not result.success:
                logger.warning(
                    "Sandbox execution failed",
                    sandbox_id=result.sandbox_id,
                    exit_code=result.exit_code,
                    timed_out=result.timed_out,
                    memory_exceeded=result.memory_exceeded,
                )

            return result

    def validate_code(self, code: str) -> tuple[bool, str | None]:
        """Basic static validation of code before execution.

        Checks for obviously dangerous patterns. This is NOT a security
        boundary — the sandbox itself provides isolation.

        Returns:
            Tuple of (is_safe, reason_if_unsafe).
        """
        dangerous_patterns = [
            ("import os; os.system", "Direct OS command execution"),
            ("subprocess.call", "Subprocess execution"),
            ("subprocess.run", "Subprocess execution"),
            ("subprocess.Popen", "Subprocess execution"),
            ("__import__('os')", "Dynamic OS import"),
            ("eval(", "Dynamic code evaluation"),
            ("exec(", "Dynamic code execution"),
            ("open('/etc/", "System file access"),
            ("open('/proc/", "Process info access"),
        ]

        for pattern, reason in dangerous_patterns:
            if pattern in code:
                return False, reason

        return True, None


__all__ = [
    "MAX_OUTPUT_BYTES",
    "ProcessSandbox",
    "Sandbox",
    "SandboxBackend",
    "SandboxConfig",
    "SandboxPolicy",
    "SandboxResult",
]



"""
ACGS-2 Security Headers Middleware
Constitutional Hash: 608508a9bd224290

Security headers middleware for FastAPI applications providing defense-in-depth
against common web attacks including XSS, clickjacking, MIME sniffing, and downgrade attacks.

Implements enterprise-grade security headers:
- Content-Security-Policy: Controls resource loading to prevent XSS
- X-Content-Type-Options: Prevents MIME sniffing attacks
- X-Frame-Options: Prevents clickjacking attacks
- Strict-Transport-Security: Enforces HTTPS connections
- X-XSS-Protection: Enables browser XSS filtering
- Referrer-Policy: Controls referrer information leakage

Environment Support:
- Development: Relaxed CSP, shorter HSTS max-age
- Staging: Moderate security, testing-friendly
- Production: Strict security headers, long HSTS max-age

Usage:
    from src.core.shared.security.security_headers import SecurityHeadersMiddleware, SecurityHeadersConfig

    # Basic usage with defaults
    app.add_middleware(SecurityHeadersMiddleware)

    # Custom configuration for production
    config = SecurityHeadersConfig(
        environment="production",
        custom_csp_directives={"connect-src": ["'self'", "wss://example.com"]}
    )
    app.add_middleware(SecurityHeadersMiddleware, config=config)

    # WebSocket-enabled configuration
    config = SecurityHeadersConfig.for_websocket_service()
    app.add_middleware(SecurityHeadersMiddleware, config=config)
"""

import os
from dataclasses import dataclass

from starlette.datastructures import MutableHeaders
from starlette.responses import Response
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from src.core.shared.config import settings
from src.core.shared.config.runtime_environment import resolve_runtime_environment
from src.core.shared.constants import CONSTITUTIONAL_HASH
from src.core.shared.structured_logging import get_logger

# Constitutional hash for governance compliance

logger = get_logger(__name__)


def _detect_environment() -> str:
    return resolve_runtime_environment(getattr(settings, "env", None))


@dataclass
class SecurityHeadersConfig:
    """
    Configuration for security headers middleware.

    Attributes:
        environment: Deployment environment (development, staging, production)
        enable_hsts: Enable Strict-Transport-Security header
        hsts_max_age: HSTS max-age in seconds
        hsts_include_subdomains: Include subdomains in HSTS
        hsts_preload: Enable HSTS preload
        custom_csp_directives: Custom Content-Security-Policy directives
        frame_options: X-Frame-Options value (DENY, SAMEORIGIN)
        referrer_policy: Referrer-Policy value
        enable_xss_protection: Enable X-XSS-Protection header
    """

    environment: str = "production"
    enable_hsts: bool = True
    hsts_max_age: int = 31536000  # 1 year in seconds
    hsts_include_subdomains: bool = True
    hsts_preload: bool = False
    custom_csp_directives: dict[str, list[str]] | None = None
    frame_options: str = "DENY"
    referrer_policy: str = "strict-origin-when-cross-origin"
    enable_xss_protection: bool = True

    @classmethod
    def from_env(cls) -> "SecurityHeadersConfig":
        """
        Create configuration from environment variables.

        Environment Variables:
            APP_ENV: Application environment (development, staging, production)
            SECURITY_HSTS_ENABLED: Enable HSTS (default: true for production)
            SECURITY_HSTS_MAX_AGE: HSTS max-age in seconds
            SECURITY_FRAME_OPTIONS: X-Frame-Options value

        Returns:
            SecurityHeadersConfig instance
        """
        environment = _detect_environment()

        # Environment-specific defaults
        if environment == "development":
            hsts_enabled = os.getenv("SECURITY_HSTS_ENABLED", "false").lower() == "true"
            hsts_max_age = int(os.getenv("SECURITY_HSTS_MAX_AGE", "300"))  # 5 minutes
        elif environment == "staging":
            hsts_enabled = os.getenv("SECURITY_HSTS_ENABLED", "true").lower() == "true"
            hsts_max_age = int(os.getenv("SECURITY_HSTS_MAX_AGE", "86400"))  # 1 day
        else:  # production
            hsts_enabled = os.getenv("SECURITY_HSTS_ENABLED", "true").lower() == "true"
            hsts_max_age = int(os.getenv("SECURITY_HSTS_MAX_AGE", "31536000"))  # 1 year

        frame_options = os.getenv("SECURITY_FRAME_OPTIONS", "DENY")

        return cls(
            environment=environment,
            enable_hsts=hsts_enabled,
            hsts_max_age=hsts_max_age,
            frame_options=frame_options,
        )

    @classmethod
    def for_development(cls) -> "SecurityHeadersConfig":
        """
        Create development-friendly configuration.

        Features:
        - Relaxed CSP allowing localhost and eval (for dev tools)
        - Short HSTS max-age (5 minutes)
        - HSTS disabled by default

        Returns:
            SecurityHeadersConfig for development
        """
        return cls(
            environment="development",
            enable_hsts=False,
            hsts_max_age=300,  # 5 minutes
            hsts_include_subdomains=False,
            custom_csp_directives={
                "default-src": ["'self'"],
                "script-src": ["'self'", "'unsafe-inline'", "'unsafe-eval'", "localhost:*"],
                "connect-src": ["'self'", "localhost:*", "ws://localhost:*"],
            },
        )

    @classmethod
    def for_production(cls, strict: bool = True) -> "SecurityHeadersConfig":
        """
        Create production-grade configuration.

        Args:
            strict: Use strict CSP settings

        Features:
        - Strict CSP with minimal allowed sources
        - Long HSTS max-age (1 year)
        - HSTS with subdomains and preload
        - Frame options DENY

        Returns:
            SecurityHeadersConfig for production
        """
        csp_directives = None
        if strict:
            csp_directives = {
                "default-src": ["'self'"],
                "script-src": ["'self'"],
                "style-src": ["'self'"],
                "img-src": ["'self'", "data:"],
                "font-src": ["'self'"],
                "connect-src": ["'self'"],
                "frame-ancestors": ["'none'"],
            }

        return cls(
            environment="production",
            enable_hsts=True,
            hsts_max_age=31536000,  # 1 year
            hsts_include_subdomains=True,
            hsts_preload=True,
            custom_csp_directives=csp_directives,
            frame_options="DENY",
        )

    @classmethod
    def for_websocket_service(cls) -> "SecurityHeadersConfig":
        """
        Create configuration for services using WebSockets.

        Features:
        - CSP allows WebSocket connections (ws:// and wss://)
        - Production-grade HSTS
        - Suitable for real-time dashboards and observability tools

        Returns:
            SecurityHeadersConfig for WebSocket services
        """
        return cls(
            environment="production",
            enable_hsts=True,
            hsts_max_age=31536000,
            custom_csp_directives={
                "default-src": ["'self'"],
                "script-src": ["'self'"],
                "style-src": ["'self'", "'unsafe-inline'"],  # Allow inline styles for UI
                "connect-src": ["'self'", "ws:", "wss:"],  # Allow WebSocket connections
                "img-src": ["'self'", "data:"],
            },
        )

    @classmethod
    def for_integration_service(cls) -> "SecurityHeadersConfig":
        """
        Create configuration for integration services.

        Features:
        - CSP allows external API connections
        - Suitable for webhook handlers and third-party integrations
        - Production-grade security with integration flexibility

        Returns:
            SecurityHeadersConfig for integration services
        """
        return cls(
            environment="production",
            enable_hsts=True,
            hsts_max_age=31536000,
            custom_csp_directives={
                "default-src": ["'self'"],
                "script-src": ["'self'"],
                "connect-src": ["'self'", "https:"],  # Allow HTTPS external connections
                "img-src": ["'self'", "data:", "https:"],
            },
        )

    def get_csp_header_value(self) -> str:
        """
        Build Content-Security-Policy header value.

        Returns:
            CSP header value string
        """
        # Default CSP directives
        default_directives = {
            "default-src": ["'self'"],
            "script-src": ["'self'"],
            "style-src": ["'self'"],
            "img-src": ["'self'", "data:"],
            "font-src": ["'self'"],
            "connect-src": ["'self'"],
            "frame-ancestors": ["'none'"],
            "base-uri": ["'self'"],
            "form-action": ["'self'"],
        }

        # Merge custom directives
        directives = default_directives.copy()
        if self.custom_csp_directives:
            directives.update(self.custom_csp_directives)

        # Build CSP string
        csp_parts = []
        for directive, sources in directives.items():
            sources_str = " ".join(sources)
            csp_parts.append(f"{directive} {sources_str}")

        return "; ".join(csp_parts)

    def get_hsts_header_value(self) -> str | None:
        """
        Build Strict-Transport-Security header value.

        Returns:
            HSTS header value string or None if disabled
        """
        if not self.enable_hsts:
            return None

        hsts_parts = [f"max-age={self.hsts_max_age}"]

        if self.hsts_include_subdomains:
            hsts_parts.append("includeSubDomains")

        if self.hsts_preload:
            hsts_parts.append("preload")

        return "; ".join(hsts_parts)


class SecurityHeadersMiddleware:
    """
    Pure ASGI middleware for adding security headers to all HTTP responses.

    ~50-80% faster than the previous BaseHTTPMiddleware implementation by
    avoiding per-request Request object cloning and coroutine boundary overhead.

    Implements:
    - Content-Security-Policy: Controls resource loading
    - X-Content-Type-Options: Prevents MIME sniffing
    - X-Frame-Options: Prevents clickjacking
    - Strict-Transport-Security: Enforces HTTPS
    - X-XSS-Protection: Enables XSS filtering
    - Referrer-Policy: Controls referrer information

    Usage:
        app.add_middleware(SecurityHeadersMiddleware)

        # With custom config
        config = SecurityHeadersConfig.for_production()
        app.add_middleware(SecurityHeadersMiddleware, config=config)
    """

    def __init__(self, app: ASGIApp, config: SecurityHeadersConfig | None = None):
        """
        Initialize security headers middleware.

        Args:
            app: FastAPI application instance
            config: SecurityHeadersConfig instance (default: from environment)
        """
        self.app = app
        self.config = config or SecurityHeadersConfig.from_env()

        logger.info(
            "Security headers middleware initialized for environment: %s (constitutional_hash=%s)",
            self.config.environment,
            CONSTITUTIONAL_HASH,
        )

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """Process ASGI request and inject security headers into response."""
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        async def send_with_security_headers(message: Message) -> None:
            if message["type"] == "http.response.start":
                headers = MutableHeaders(scope=message)
                self._set_security_headers(headers)
            await send(message)

        await self.app(scope, receive, send_with_security_headers)

    def _set_security_headers(self, headers: MutableHeaders) -> None:
        """
        Set all security headers on a MutableHeaders-compatible object.

        Args:
            headers: MutableHeaders (or Response.headers) to modify
        """
        # Content-Security-Policy
        headers["Content-Security-Policy"] = self.config.get_csp_header_value()

        # X-Content-Type-Options
        headers["X-Content-Type-Options"] = "nosniff"

        # X-Frame-Options
        headers["X-Frame-Options"] = self.config.frame_options

        # Strict-Transport-Security (if enabled)
        hsts_value = self.config.get_hsts_header_value()
        if hsts_value:
            headers["Strict-Transport-Security"] = hsts_value

        # X-XSS-Protection
        if self.config.enable_xss_protection:
            headers["X-XSS-Protection"] = "1; mode=block"

        # Referrer-Policy
        headers["Referrer-Policy"] = self.config.referrer_policy

        # Permissions-Policy — restrict unnecessary browser features for API surfaces
        headers["Permissions-Policy"] = "camera=(), microphone=(), geolocation=(), payment=()"

    def _add_security_headers(self, response: Response) -> None:
        """
        Add all security headers to a Response object.

        Backward-compatible helper used by direct callers.

        Args:
            response: HTTP response to modify
        """
        self._set_security_headers(response.headers)


def add_security_headers(
    app: ASGIApp,
    environment: str | None = None,
    config: SecurityHeadersConfig | None = None,
) -> None:
    """
    Convenience function to add security headers middleware to a FastAPI app.

    Args:
        app: FastAPI application instance
        environment: Environment name (development, staging, production)
        config: Custom SecurityHeadersConfig (overrides environment)

    Example:
        from src.core.shared.security.security_headers import add_security_headers

        app = FastAPI()
        add_security_headers(app, environment="production")
    """
    if config is None:
        if environment == "development":
            config = SecurityHeadersConfig.for_development()
        elif environment == "production":
            config = SecurityHeadersConfig.for_production()
        else:
            config = SecurityHeadersConfig.from_env()

    app.add_middleware(SecurityHeadersMiddleware, config=config)

    logger.info(
        "Added security headers middleware with environment: %s (constitutional_hash=%s)",
        config.environment,
        CONSTITUTIONAL_HASH,
    )



"""Comprehensive tests for OIDC handler.

Constitutional Hash: 608508a9bd224290

Covers:
- Dataclass construction and validation (OIDCProviderConfig, OIDCTokenResponse, OIDCUserInfo)
- OIDCHandler registration, lookup, listing
- Async flows: initiate_login, handle_callback, refresh_token, logout
- PKCE code challenge generation
- Pending state management and eviction (H-4 DoS fix)
- Error paths: missing provider, invalid state, provider mismatch, network errors
- Edge cases: empty claims, Azure AD group claims, metadata caching
"""

from __future__ import annotations

import base64
import hashlib
from datetime import UTC, datetime, timedelta
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from src.core.shared.auth.oidc_handler import (
    DEFAULT_SCOPES,
    OIDCAuthenticationError,
    OIDCConfigurationError,
    OIDCError,
    OIDCHandler,
    OIDCProviderConfig,
    OIDCProviderError,
    OIDCTokenError,
    OIDCTokenResponse,
    OIDCUserInfo,
    _normalize_secret_sentinel,
)

# ---------------------------------------------------------------------------
# Fixtures
# ---------------------------------------------------------------------------

MOCK_METADATA: dict = {
    "issuer": "https://accounts.google.com",
    "authorization_endpoint": "https://accounts.google.com/o/oauth2/v2/auth",
    "token_endpoint": "https://oauth2.googleapis.com/token",
    "userinfo_endpoint": "https://openidconnect.googleapis.com/v1/userinfo",
    "jwks_uri": "https://www.googleapis.com/oauth2/v3/certs",
    "end_session_endpoint": "https://accounts.google.com/logout",
}

VALID_PROVIDER_KWARGS: dict = {
    "name": "google",
    "client_id": "test-client-id",
    "client_secret": "real-production-secret-abc123",
    "server_metadata_url": "https://accounts.google.com/.well-known/openid-configuration",
}


@pytest.fixture()
def handler() -> OIDCHandler:
    return OIDCHandler()


@pytest.fixture()
def registered_handler(handler: OIDCHandler) -> OIDCHandler:
    handler.register_provider(**VALID_PROVIDER_KWARGS)
    return handler


def _mock_http_response(*, status_code: int = 200, json_data: dict | None = None) -> MagicMock:
    resp = MagicMock()
    resp.status_code = status_code
    resp.content = b'{"ok":true}' if json_data else b""
    resp.json.return_value = json_data or {}
    resp.raise_for_status = MagicMock()
    if status_code >= 400:
        import httpx

        resp.raise_for_status.side_effect = httpx.HTTPStatusError(
            "error", request=MagicMock(), response=resp
        )
    return resp


# ---------------------------------------------------------------------------
# _normalize_secret_sentinel
# ---------------------------------------------------------------------------


class TestNormalizeSecretSentinel:
    def test_strips_and_lowercases(self):
        assert _normalize_secret_sentinel("  Replace-Me ") == "replaceme"

    def test_removes_non_alnum(self):
        assert _normalize_secret_sentinel("your--secret!!") == "yoursecret"

    def test_empty_string(self):
        assert _normalize_secret_sentinel("") == ""


# ---------------------------------------------------------------------------
# OIDCError hierarchy
# ---------------------------------------------------------------------------


class TestOIDCErrors:
    def test_base_error_status(self):
        assert OIDCError.http_status_code == 500

    def test_config_error_status(self):
        assert OIDCConfigurationError.http_status_code == 500

    def test_auth_error_status(self):
        assert OIDCAuthenticationError.http_status_code == 401

    def test_token_error_status(self):
        assert OIDCTokenError.http_status_code == 401

    def test_provider_error_status(self):
        assert OIDCProviderError.http_status_code == 502

    def test_inheritance(self):
        err = OIDCConfigurationError("bad config")
        assert isinstance(err, OIDCError)


# ---------------------------------------------------------------------------
# OIDCProviderConfig
# ---------------------------------------------------------------------------


class TestOIDCProviderConfig:
    def test_valid_config(self):
        config = OIDCProviderConfig(**VALID_PROVIDER_KWARGS)
        assert config.name == "google"
        assert config.use_pkce is True
        assert config.scopes == list(DEFAULT_SCOPES)
        assert config.extra_params == {}

    def test_custom_scopes(self):
        config = OIDCProviderConfig(**{**VALID_PROVIDER_KWARGS, "scopes": ["openid", "custom"]})
        assert config.scopes == ["openid", "custom"]

    def test_empty_name_rejected(self):
        with pytest.raises(OIDCConfigurationError, match="Provider name is required"):
            OIDCProviderConfig(
                name="",
                client_id="cid",
                client_secret="secret-value-ok",
                server_metadata_url="https://example.com",
            )

    def test_empty_client_id_rejected(self):
        with pytest.raises(OIDCConfigurationError, match="Client ID is required"):
            OIDCProviderConfig(
                name="test",
                client_id="",
                client_secret="secret-value-ok",
                server_metadata_url="https://example.com",
            )

    def test_empty_client_secret_rejected(self):
        with pytest.raises(OIDCConfigurationError, match="Client secret is required"):
            OIDCProviderConfig(
                name="test",
                client_id="cid",
                client_secret="",
                server_metadata_url="https://example.com",
            )

    def test_empty_metadata_url_rejected(self):
        with pytest.raises(OIDCConfigurationError, match="Server metadata URL is required"):
            OIDCProviderConfig(
                name="test",
                client_id="cid",
                client_secret="secret-value-ok",
                server_metadata_url="",
            )

    def test_placeholder_replace_me_rejected(self):
        with pytest.raises(OIDCConfigurationError, match="placeholder value"):
            OIDCProviderConfig(
                name="t",
                client_id="c",
                client_secret="REPLACE_ME",  # pragma: allowlist secret
                server_metadata_url="https://x.com",
            )

    def test_placeholder_your_secret_rejected(self):
        with pytest.raises(OIDCConfigurationError, match="placeholder value"):
            OIDCProviderConfig(
                name="t",
                client_id="c",
                client_secret="your-secret",  # pragma: allowlist secret
                server_metadata_url="https://x.com",
            )

    def test_placeholder_variant_with_spaces(self):
        with pytest.raises(OIDCConfigurationError, match="placeholder value"):
            OIDCProviderConfig(
                name="t",
                client_id="c",
                client_secret=" replace-me ",
                server_metadata_url="https://x.com",
            )


# ---------------------------------------------------------------------------
# OIDCTokenResponse
# ---------------------------------------------------------------------------


class TestOIDCTokenResponse:
    def test_from_dict_full(self):
        data = {
            "access_token": "at-123",
            "token_type": "Bearer",
            "expires_in": 3600,
            "refresh_token": "rt-456",
            "id_token": "idt-789",
            "scope": "openid profile",
        }
        resp = OIDCTokenResponse.from_dict(data)
        assert resp.access_token == "at-123"
        assert resp.token_type == "Bearer"
        assert resp.expires_in == 3600
        assert resp.refresh_token == "rt-456"
        assert resp.id_token == "idt-789"
        assert resp.scope == "openid profile"
        assert resp.raw_response is data

    def test_from_dict_minimal(self):
        data = {"access_token": "at"}
        resp = OIDCTokenResponse.from_dict(data)
        assert resp.access_token == "at"
        assert resp.token_type == "Bearer"
        assert resp.expires_in is None
        assert resp.refresh_token is None
        assert resp.id_token is None
        assert resp.scope is None

    def test_from_dict_empty_access_token(self):
        resp = OIDCTokenResponse.from_dict({})
        assert resp.access_token == ""


# ---------------------------------------------------------------------------
# OIDCUserInfo
# ---------------------------------------------------------------------------


class TestOIDCUserInfo:
    def test_from_claims_full(self):
        claims = {
            "sub": "user-1",
            "email": "u@e.com",
            "email_verified": True,
            "name": "User One",
            "given_name": "User",
            "family_name": "One",
            "picture": "https://pic.png",
            "locale": "en",
            "groups": ["admin"],
        }
        info = OIDCUserInfo.from_claims(claims)
        assert info.sub == "user-1"
        assert info.email == "u@e.com"
        assert info.email_verified is True
        assert info.name == "User One"
        assert info.groups == ["admin"]
        assert info.raw_claims is claims

    def test_from_claims_minimal(self):
        info = OIDCUserInfo.from_claims({"sub": "x"})
        assert info.sub == "x"
        assert info.email is None
        assert info.groups == []

    def test_from_claims_roles_fallback(self):
        info = OIDCUserInfo.from_claims({"sub": "x", "roles": ["editor"]})
        assert info.groups == ["editor"]

    def test_from_claims_azure_ad_groups(self):
        info = OIDCUserInfo.from_claims(
            {
                "sub": "x",
                "https://schemas.microsoft.com/claims/groups": ["g1", "g2"],
            }
        )
        assert info.groups == ["g1", "g2"]

    def test_from_claims_non_list_groups_ignored(self):
        info = OIDCUserInfo.from_claims({"sub": "x", "groups": "not-a-list"})
        assert info.groups == []

    def test_from_claims_empty(self):
        info = OIDCUserInfo.from_claims({})
        assert info.sub == ""


# ---------------------------------------------------------------------------
# OIDCHandler — sync methods
# ---------------------------------------------------------------------------


class TestOIDCHandlerSync:
    def test_init(self, handler: OIDCHandler):
        assert handler.list_providers() == []

    def test_register_provider(self, handler: OIDCHandler):
        handler.register_provider(**VALID_PROVIDER_KWARGS)
        assert "google" in handler.list_providers()

    def test_register_multiple_providers(self, handler: OIDCHandler):
        handler.register_provider(**VALID_PROVIDER_KWARGS)
        handler.register_provider(**{**VALID_PROVIDER_KWARGS, "name": "azure"})
        assert sorted(handler.list_providers()) == ["azure", "google"]

    def test_get_provider_success(self, registered_handler: OIDCHandler):
        config = registered_handler.get_provider("google")
        assert config.name == "google"
        assert config.client_id == "test-client-id"

    def test_get_provider_not_found(self, handler: OIDCHandler):
        with pytest.raises(OIDCConfigurationError, match="not registered"):
            handler.get_provider("missing")

    def test_validate_state_false(self, handler: OIDCHandler):
        assert handler.validate_state("nonexistent") is False

    def test_generate_state_is_url_safe(self, handler: OIDCHandler):
        state = handler._generate_state()
        assert len(state) > 20
        # URL-safe base64 chars only
        assert all(c.isalnum() or c in "-_" for c in state)

    def test_generate_code_verifier(self, handler: OIDCHandler):
        verifier = handler._generate_code_verifier()
        assert len(verifier) > 40

    def test_generate_code_challenge_s256(self, handler: OIDCHandler):
        verifier = "test-verifier-value"
        challenge = handler._generate_code_challenge(verifier)
        # Verify it matches expected SHA256 + base64url
        expected_digest = hashlib.sha256(verifier.encode()).digest()
        expected = base64.urlsafe_b64encode(expected_digest).rstrip(b"=").decode("ascii")
        assert challenge == expected

    def test_clear_expired_states_removes_old(self, handler: OIDCHandler):
        old_time = (datetime.now(UTC) - timedelta(seconds=700)).isoformat()
        handler._pending_states["old-state"] = {
            "provider": "test",
            "redirect_uri": "https://cb",
            "code_verifier": None,
            "nonce": "n",
            "created_at": old_time,
        }
        handler._pending_states["fresh-state"] = {
            "provider": "test",
            "redirect_uri": "https://cb",
            "code_verifier": None,
            "nonce": "n",
            "created_at": datetime.now(UTC).isoformat(),
        }
        removed = handler.clear_expired_states(max_age_seconds=600)
        assert removed == 1
        assert "old-state" not in handler._pending_states
        assert "fresh-state" in handler._pending_states

    def test_clear_expired_states_returns_zero_when_none_expired(self, handler: OIDCHandler):
        assert handler.clear_expired_states() == 0

    def test_evict_stale_pending_states(self, handler: OIDCHandler):
        old_time = (datetime.now(UTC) - timedelta(seconds=700)).isoformat()
        handler._pending_states["stale"] = {
            "provider": "p",
            "redirect_uri": "u",
            "code_verifier": None,
            "nonce": "n",
            "created_at": old_time,
        }
        handler._evict_stale_pending_states()
        assert "stale" not in handler._pending_states


# ---------------------------------------------------------------------------
# OIDCHandler — register_provider_from_model
# ---------------------------------------------------------------------------


class TestRegisterProviderFromModel:
    def test_non_oidc_provider_rejected(self, handler: OIDCHandler):
        model = MagicMock()
        model.is_oidc = False
        model.name = "saml-only"
        with pytest.raises(OIDCConfigurationError, match="not an OIDC provider"):
            handler.register_provider_from_model(model)

    def test_invalid_config_rejected(self, handler: OIDCHandler):
        model = MagicMock()
        model.is_oidc = True
        model.name = "bad"
        model.validate_oidc_config.return_value = ["missing client_id"]
        with pytest.raises(OIDCConfigurationError, match="missing client_id"):
            handler.register_provider_from_model(model)

    def test_valid_model_registers(self, handler: OIDCHandler):
        model = MagicMock()
        model.is_oidc = True
        model.name = "okta"
        model.validate_oidc_config.return_value = []
        model.oidc_client_id = "cid"
        model.oidc_client_secret = "real-secret-value-xyz"
        model.oidc_metadata_url = "https://okta.example.com/.well-known/openid-configuration"
        model.oidc_scope_list = ["openid", "profile"]
        model.get_config.return_value = {"prompt": "consent"}
        handler.register_provider_from_model(model)
        assert "okta" in handler.list_providers()


# ---------------------------------------------------------------------------
# OIDCHandler — async methods
# ---------------------------------------------------------------------------


@pytest.mark.asyncio
class TestOIDCHandlerAsync:
    async def test_initiate_login_builds_url(self, registered_handler: OIDCHandler):
        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=_mock_http_response(json_data=MOCK_METADATA))
        registered_handler._http_client = mock_client

        auth_url, state = await registered_handler.initiate_login(
            "google", "https://app.test/callback"
        )

        assert "https://accounts.google.com/o/oauth2/v2/auth?" in auth_url
        assert "client_id=test-client-id" in auth_url
        assert "response_type=code" in auth_url
        assert "redirect_uri=https" in auth_url
        assert "state=" in auth_url
        assert "code_challenge=" in auth_url
        assert "code_challenge_method=S256" in auth_url
        assert len(state) > 20
        assert registered_handler.validate_state(state) is True

    async def test_initiate_login_without_pkce(self, handler: OIDCHandler):
        handler.register_provider(**{**VALID_PROVIDER_KWARGS, "use_pkce": False})
        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=_mock_http_response(json_data=MOCK_METADATA))
        handler._http_client = mock_client

        auth_url, state = await handler.initiate_login("google", "https://cb")
        assert "code_challenge" not in auth_url
        registered_state = handler._pending_states.get(state)
        assert registered_state is not None
        assert registered_state["code_verifier"] is None

    async def test_initiate_login_unknown_provider(self, handler: OIDCHandler):
        with pytest.raises(OIDCConfigurationError, match="not registered"):
            await handler.initiate_login("unknown", "https://cb")

    async def test_initiate_login_no_auth_endpoint(self, registered_handler: OIDCHandler):
        meta_no_auth = {k: v for k, v in MOCK_METADATA.items() if k != "authorization_endpoint"}
        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=_mock_http_response(json_data=meta_no_auth))
        registered_handler._http_client = mock_client

        with pytest.raises(OIDCProviderError, match="Authorization endpoint not found"):
            await registered_handler.initiate_login("google", "https://cb")

    async def test_initiate_login_evicts_at_capacity(self, registered_handler: OIDCHandler):
        """H-4: when pending states hit max, oldest is evicted."""
        registered_handler._max_pending_states = 2
        now = datetime.now(UTC).isoformat()
        registered_handler._pending_states["oldest"] = {
            "provider": "google",
            "redirect_uri": "https://cb",
            "code_verifier": None,
            "nonce": "n",
            "created_at": now,
        }
        registered_handler._pending_states["second"] = {
            "provider": "google",
            "redirect_uri": "https://cb",
            "code_verifier": None,
            "nonce": "n",
            "created_at": now,
        }
        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=_mock_http_response(json_data=MOCK_METADATA))
        registered_handler._http_client = mock_client

        _, state = await registered_handler.initiate_login("google", "https://cb")
        # oldest should have been evicted to make room
        assert "oldest" not in registered_handler._pending_states
        assert state in registered_handler._pending_states

    async def test_handle_callback_invalid_state(self, registered_handler: OIDCHandler):
        with pytest.raises(OIDCAuthenticationError, match="Invalid or expired state"):
            await registered_handler.handle_callback("google", "code", "bad-state")

    async def test_handle_callback_provider_mismatch(self, registered_handler: OIDCHandler):
        state = "test-state-123"
        registered_handler._pending_states[state] = {
            "provider": "azure",
            "redirect_uri": "https://cb",
            "code_verifier": None,
            "nonce": "nonce",
            "created_at": datetime.now(UTC).isoformat(),
        }
        with pytest.raises(OIDCAuthenticationError, match="Provider mismatch"):
            await registered_handler.handle_callback("google", "code", state)

    async def test_handle_callback_success(self, registered_handler: OIDCHandler):
        """Full callback flow: state validation -> code exchange -> user info."""
        state = "valid-state"
        registered_handler._pending_states[state] = {
            "provider": "google",
            "redirect_uri": "https://app/callback",
            "code_verifier": "verifier-abc",
            "nonce": "nonce-123",
            "created_at": datetime.now(UTC).isoformat(),
        }

        token_response_data = {
            "access_token": "at-ok",
            "token_type": "Bearer",
            "expires_in": 3600,
            "id_token": None,
        }
        userinfo_data = {
            "sub": "u-42",
            "email": "u@example.com",
            "email_verified": True,
            "name": "Test",
        }

        mock_client = AsyncMock()

        async def mock_get(url, **kwargs):
            if "well-known" in url or "openid-configuration" in url:
                return _mock_http_response(json_data=MOCK_METADATA)
            if "userinfo" in url:
                return _mock_http_response(json_data=userinfo_data)
            return _mock_http_response(json_data=MOCK_METADATA)

        async def mock_post(url, **kwargs):
            return _mock_http_response(json_data=token_response_data)

        mock_client.get = mock_get
        mock_client.post = mock_post
        registered_handler._http_client = mock_client

        user = await registered_handler.handle_callback("google", "auth-code", state)

        assert user.sub == "u-42"
        assert user.email == "u@example.com"
        # State should be consumed
        assert state not in registered_handler._pending_states

    async def test_handle_callback_uses_redirect_uri_param(self, registered_handler: OIDCHandler):
        """When redirect_uri is passed explicitly it takes precedence."""
        state = "s1"
        registered_handler._pending_states[state] = {
            "provider": "google",
            "redirect_uri": "https://stored-uri/cb",
            "code_verifier": None,
            "nonce": "n",
            "created_at": datetime.now(UTC).isoformat(),
        }

        captured_data: dict = {}

        async def mock_post(url, data=None, **kwargs):
            captured_data.update(data or {})
            return _mock_http_response(json_data={"access_token": "at"})

        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=_mock_http_response(json_data=MOCK_METADATA))
        mock_client.post = mock_post
        registered_handler._http_client = mock_client

        # Patch _get_user_info to shortcut the full flow
        registered_handler._get_user_info = AsyncMock(return_value=OIDCUserInfo(sub="u1"))

        await registered_handler.handle_callback(
            "google", "code", state, redirect_uri="https://override/cb"
        )
        assert captured_data.get("redirect_uri") == "https://override/cb"

    async def test_exchange_code_no_token_endpoint(self, registered_handler: OIDCHandler):
        meta_no_token = {k: v for k, v in MOCK_METADATA.items() if k != "token_endpoint"}
        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=_mock_http_response(json_data=meta_no_token))
        registered_handler._http_client = mock_client
        provider = registered_handler.get_provider("google")

        with pytest.raises(OIDCTokenError, match="Token endpoint not found"):
            await registered_handler._exchange_code(provider, "code", "https://cb")

    async def test_exchange_code_non_200(self, registered_handler: OIDCHandler):
        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=_mock_http_response(json_data=MOCK_METADATA))
        error_resp = _mock_http_response(
            status_code=400,
            json_data={"error": "invalid_grant", "error_description": "Code expired"},
        )
        # Override raise_for_status to not raise (exchange_code checks status_code directly)
        error_resp.raise_for_status = MagicMock()
        mock_client.post = AsyncMock(return_value=error_resp)
        registered_handler._http_client = mock_client
        provider = registered_handler.get_provider("google")

        with pytest.raises(OIDCTokenError, match="Code expired"):
            await registered_handler._exchange_code(provider, "code", "https://cb")

    async def test_exchange_code_network_error(self, registered_handler: OIDCHandler):
        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=_mock_http_response(json_data=MOCK_METADATA))
        mock_client.post = AsyncMock(side_effect=RuntimeError("connection reset"))
        registered_handler._http_client = mock_client
        provider = registered_handler.get_provider("google")

        with pytest.raises(OIDCTokenError, match="connection reset"):
            await registered_handler._exchange_code(provider, "code", "https://cb")

    async def test_refresh_token_success(self, registered_handler: OIDCHandler):
        token_data = {"access_token": "new-at", "expires_in": 7200}
        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=_mock_http_response(json_data=MOCK_METADATA))
        mock_client.post = AsyncMock(return_value=_mock_http_response(json_data=token_data))
        registered_handler._http_client = mock_client

        result = await registered_handler.refresh_token("google", "old-rt")
        assert result.access_token == "new-at"
        assert result.expires_in == 7200

    async def test_refresh_token_no_endpoint(self, registered_handler: OIDCHandler):
        meta_no_token = {k: v for k, v in MOCK_METADATA.items() if k != "token_endpoint"}
        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=_mock_http_response(json_data=meta_no_token))
        registered_handler._http_client = mock_client

        with pytest.raises(OIDCTokenError, match="Token endpoint not found"):
            await registered_handler.refresh_token("google", "rt")

    async def test_refresh_token_non_200(self, registered_handler: OIDCHandler):
        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=_mock_http_response(json_data=MOCK_METADATA))
        error_resp = _mock_http_response(status_code=400, json_data={"error": "invalid_grant"})
        error_resp.raise_for_status = MagicMock()
        mock_client.post = AsyncMock(return_value=error_resp)
        registered_handler._http_client = mock_client

        with pytest.raises(OIDCTokenError, match="Token refresh failed"):
            await registered_handler.refresh_token("google", "rt")

    async def test_refresh_token_unknown_provider(self, handler: OIDCHandler):
        with pytest.raises(OIDCConfigurationError, match="not registered"):
            await handler.refresh_token("missing", "rt")

    async def test_logout_with_session_endpoint(self, registered_handler: OIDCHandler):
        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=_mock_http_response(json_data=MOCK_METADATA))
        registered_handler._http_client = mock_client

        url = await registered_handler.logout(
            "google",
            id_token_hint="idt-hint",
            post_logout_redirect_uri="https://app/logged-out",
        )
        assert url is not None
        assert "https://accounts.google.com/logout?" in url
        assert "id_token_hint=idt-hint" in url
        assert "post_logout_redirect_uri=" in url
        assert "client_id=test-client-id" in url

    async def test_logout_no_session_endpoint(self, registered_handler: OIDCHandler):
        meta_no_logout = {k: v for k, v in MOCK_METADATA.items() if k != "end_session_endpoint"}
        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=_mock_http_response(json_data=meta_no_logout))
        registered_handler._http_client = mock_client

        url = await registered_handler.logout("google")
        assert url is None

    async def test_logout_minimal_params(self, registered_handler: OIDCHandler):
        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=_mock_http_response(json_data=MOCK_METADATA))
        registered_handler._http_client = mock_client

        url = await registered_handler.logout("google")
        assert url is not None
        assert "id_token_hint" not in url

    async def test_close(self, handler: OIDCHandler):
        mock_client = AsyncMock()
        handler._http_client = mock_client
        await handler.close()
        mock_client.aclose.assert_awaited_once()
        assert handler._http_client is None

    async def test_close_when_no_client(self, handler: OIDCHandler):
        # Should not raise
        await handler.close()
        assert handler._http_client is None

    async def test_fetch_metadata_caching(self, registered_handler: OIDCHandler):
        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=_mock_http_response(json_data=MOCK_METADATA))
        registered_handler._http_client = mock_client
        provider = registered_handler.get_provider("google")

        # First call fetches
        meta1 = await registered_handler._fetch_metadata(provider)
        assert meta1["issuer"] == "https://accounts.google.com"
        assert mock_client.get.await_count == 1

        # Second call uses cache
        meta2 = await registered_handler._fetch_metadata(provider)
        assert meta2 == meta1
        assert mock_client.get.await_count == 1  # no additional call

    async def test_fetch_metadata_force_refresh(self, registered_handler: OIDCHandler):
        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=_mock_http_response(json_data=MOCK_METADATA))
        registered_handler._http_client = mock_client
        provider = registered_handler.get_provider("google")

        await registered_handler._fetch_metadata(provider)
        await registered_handler._fetch_metadata(provider, force_refresh=True)
        assert mock_client.get.await_count == 2

    async def test_fetch_metadata_error_uses_cache(self, registered_handler: OIDCHandler):
        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=_mock_http_response(json_data=MOCK_METADATA))
        registered_handler._http_client = mock_client
        provider = registered_handler.get_provider("google")

        # Prime the cache
        await registered_handler._fetch_metadata(provider)

        # Next fetch fails but cache is available
        mock_client.get = AsyncMock(side_effect=RuntimeError("network down"))
        meta = await registered_handler._fetch_metadata(provider, force_refresh=True)
        assert meta["issuer"] == "https://accounts.google.com"

    async def test_fetch_metadata_error_no_cache_raises(self, registered_handler: OIDCHandler):
        mock_client = AsyncMock()
        mock_client.get = AsyncMock(side_effect=RuntimeError("network down"))
        registered_handler._http_client = mock_client
        provider = registered_handler.get_provider("google")

        with pytest.raises(OIDCProviderError, match="Failed to fetch OIDC metadata"):
            await registered_handler._fetch_metadata(provider)

    async def test_decode_id_token_missing_jwks_uri_precedes_authlib_requirement(
        self, registered_handler: OIDCHandler
    ):
        provider = registered_handler.get_provider("google")
        registered_handler._fetch_metadata = AsyncMock(
            return_value={"issuer": MOCK_METADATA["issuer"]}
        )

        with patch("src.core.shared.auth.oidc_handler.HAS_AUTHLIB", False):
            with pytest.raises(OIDCTokenError, match="JWKS URI not found"):
                await registered_handler._decode_id_token("token", provider)

    async def test_decode_id_token_no_authlib_with_valid_metadata(
        self, registered_handler: OIDCHandler
    ):
        provider = registered_handler.get_provider("google")
        registered_handler._fetch_metadata = AsyncMock(return_value=MOCK_METADATA)

        with patch("src.core.shared.auth.oidc_handler.HAS_AUTHLIB", False):
            with pytest.raises(OIDCTokenError, match="authlib library"):
                await registered_handler._decode_id_token("token", provider)

    async def test_fetch_userinfo_success(self, registered_handler: OIDCHandler):
        userinfo = {"sub": "u1", "email": "a@b.com"}
        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=_mock_http_response(json_data=MOCK_METADATA))
        registered_handler._http_client = mock_client
        provider = registered_handler.get_provider("google")

        # Prime metadata cache
        await registered_handler._fetch_metadata(provider)

        # Now mock the userinfo call
        mock_client.get = AsyncMock(return_value=_mock_http_response(json_data=userinfo))
        result = await registered_handler._fetch_userinfo(provider, "access-token")
        assert result.sub == "u1"
        assert result.email == "a@b.com"

    async def test_fetch_userinfo_no_endpoint(self, registered_handler: OIDCHandler):
        meta_no_ui = {k: v for k, v in MOCK_METADATA.items() if k != "userinfo_endpoint"}
        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=_mock_http_response(json_data=meta_no_ui))
        registered_handler._http_client = mock_client
        provider = registered_handler.get_provider("google")

        with pytest.raises(OIDCProviderError, match="Userinfo endpoint not found"):
            await registered_handler._fetch_userinfo(provider, "at")

    async def test_fetch_userinfo_non_200(self, registered_handler: OIDCHandler):
        mock_client = AsyncMock()
        # First call: metadata (cached already or fetched fresh)
        registered_handler._metadata_cache["google"] = MOCK_METADATA
        registered_handler._metadata_timestamps["google"] = datetime.now(UTC)

        error_resp = _mock_http_response(status_code=401, json_data={})
        error_resp.raise_for_status = MagicMock()
        mock_client.get = AsyncMock(return_value=error_resp)
        registered_handler._http_client = mock_client
        provider = registered_handler.get_provider("google")

        with pytest.raises(OIDCProviderError, match="Userinfo request failed"):
            await registered_handler._fetch_userinfo(provider, "bad-at")

    async def test_get_user_info_fallback_to_userinfo_endpoint(
        self, registered_handler: OIDCHandler
    ):
        """When id_token decoding fails, falls back to userinfo endpoint."""
        registered_handler._metadata_cache["google"] = MOCK_METADATA
        registered_handler._metadata_timestamps["google"] = datetime.now(UTC)

        tokens = OIDCTokenResponse(
            access_token="at",
            id_token="invalid-jwt",
        )
        userinfo_data = {"sub": "fallback-user", "email": "fb@x.com"}

        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=_mock_http_response(json_data=userinfo_data))
        registered_handler._http_client = mock_client
        provider = registered_handler.get_provider("google")

        # _decode_id_token will fail since we don't have real keys
        with patch.object(
            registered_handler,
            "_decode_id_token",
            new_callable=AsyncMock,
            side_effect=OIDCTokenError("invalid token"),
        ):
            result = await registered_handler._get_user_info(provider, tokens)

        assert result.sub == "fallback-user"

    async def test_get_user_info_from_id_token(self, registered_handler: OIDCHandler):
        """When id_token decoding succeeds, returns claims directly."""
        tokens = OIDCTokenResponse(access_token="at", id_token="valid-jwt")
        provider = registered_handler.get_provider("google")

        claims = {"sub": "jwt-user", "email": "jwt@x.com", "email_verified": True}
        with patch.object(
            registered_handler,
            "_decode_id_token",
            new_callable=AsyncMock,
            return_value=claims,
        ):
            result = await registered_handler._get_user_info(provider, tokens)

        assert result.sub == "jwt-user"

    async def test_get_user_info_no_id_token(self, registered_handler: OIDCHandler):
        """When no id_token, goes straight to userinfo endpoint."""
        registered_handler._metadata_cache["google"] = MOCK_METADATA
        registered_handler._metadata_timestamps["google"] = datetime.now(UTC)

        tokens = OIDCTokenResponse(access_token="at", id_token=None)
        userinfo_data = {"sub": "endpoint-user"}

        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=_mock_http_response(json_data=userinfo_data))
        registered_handler._http_client = mock_client
        provider = registered_handler.get_provider("google")

        result = await registered_handler._get_user_info(provider, tokens)
        assert result.sub == "endpoint-user"

    async def test_get_http_client_creates_once(self, handler: OIDCHandler):
        client = await handler._get_http_client()
        assert client is not None
        client2 = await handler._get_http_client()
        assert client is client2
        # Cleanup
        await handler.close()



"""
Tests for decisions.py route coverage.
Constitutional Hash: 608508a9bd224290

Covers: get_decision_explanation, generate_decision_explanation,
        get_governance_vector_schema, _convert_to_response helper.
"""

from __future__ import annotations

from datetime import UTC, datetime
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient

from src.core.services.api_gateway.routes.decisions import (
    _convert_to_response,
    decisions_v1_router,
)
from src.core.shared.security.auth import UserClaims, get_current_user

# ---------------------------------------------------------------------------
# Fixtures
# ---------------------------------------------------------------------------


def _make_user() -> UserClaims:
    now = int(datetime.now(UTC).timestamp())
    return UserClaims(
        sub="user-1",
        tenant_id="tenant-1",
        roles=["user"],
        permissions=["read"],
        exp=now + 3600,
        iat=now,
    )


_USER = _make_user()


def _build_app() -> FastAPI:
    app = FastAPI()
    app.include_router(decisions_v1_router)
    app.dependency_overrides[get_current_user] = lambda: _USER
    return app


@pytest.fixture()
def client() -> TestClient:
    return TestClient(_build_app())


def _mock_factor(*, governance_dim_has_value: bool = True):
    f = MagicMock()
    f.factor_id = "f-1"
    f.factor_name = "safety_score"
    f.factor_value = 0.8
    f.factor_weight = 0.5
    f.explanation = "High safety impact"
    f.evidence = ["evidence-1"]
    if governance_dim_has_value:
        f.governance_dimension = MagicMock(value="safety")
    else:
        # Use a plain string (no .value attribute) to test the str() fallback
        f.governance_dimension = "safety"
    f.source_component = "safety-engine"
    f.calculation_method = "weighted-sum"
    return f


def _mock_counterfactual(*, outcome_has_value: bool = True):
    cf = MagicMock()
    cf.scenario_id = "cf-1"
    cf.scenario_description = "Lower safety"
    cf.modified_factor = "safety_score"
    cf.original_value = 0.8
    cf.modified_value = 0.3
    if outcome_has_value:
        cf.predicted_outcome = MagicMock(value="DENY")
    else:
        # Use a plain string (no .value attribute) to test the str() fallback
        cf.predicted_outcome = "DENY"
    cf.confidence = 0.9
    cf.threshold_crossed = "safety_threshold"
    cf.impact_delta = -0.5
    return cf


def _mock_euaiact_info():
    info = MagicMock()
    info.article_13_compliant = True
    info.human_oversight_level = "human-on-the-loop"
    info.risk_category = "limited"
    info.transparency_measures = ["logging"]
    info.data_governance_info = {}
    info.technical_documentation_ref = "doc-ref-1"
    info.conformity_assessment_status = "pending"
    info.intended_purpose = "governance"
    info.limitations_and_risks = ["model drift"]
    info.human_reviewers = ["admin"]
    return info


def _mock_explanation(*, include_factors: bool = True, include_cf: bool = True):
    exp = MagicMock()
    exp.decision_id = "dec-1"
    exp.message_id = "msg-1"
    exp.request_id = "req-1"
    exp.verdict = "ALLOW"
    exp.confidence_score = 0.95
    exp.impact_score = 0.2
    exp.governance_vector = {
        "safety": 0.8,
        "security": 0.5,
        "privacy": 0.3,
        "fairness": 0.4,
        "reliability": 0.7,
        "transparency": 0.6,
        "efficiency": 0.5,
    }
    exp.factors = [_mock_factor()] if include_factors else []
    exp.primary_factors = ["f-1"] if include_factors else []
    exp.counterfactual_hints = [_mock_counterfactual()] if include_cf else []
    exp.counterfactuals_generated = include_cf
    exp.matched_rules = ["rule-1"]
    exp.violated_rules = []
    exp.applicable_policies = ["policy-1"]
    exp.summary = "Decision allowed"
    exp.detailed_reasoning = "All factors passed"
    exp.euaiact_article13_info = _mock_euaiact_info()
    exp.processing_time_ms = 12.5
    exp.explanation_generated_at = datetime.now(UTC)
    exp.explanation_version = "v1.0.0"
    exp.tenant_id = "tenant-1"
    exp.scope = "decision"
    exp.audit_references = ["audit-1"]
    exp.constitutional_hash = "608508a9bd224290"
    return exp


# ---------------------------------------------------------------------------
# _convert_to_response unit tests
# ---------------------------------------------------------------------------


class TestConvertToResponse:
    """Unit tests for the _convert_to_response helper."""

    def test_convert_full_explanation(self):
        exp = _mock_explanation()
        result = _convert_to_response(exp, include_counterfactuals=True)
        assert result.decision_id == "dec-1"
        assert result.verdict == "ALLOW"
        assert len(result.factors) == 1
        assert result.factors[0].governance_dimension == "safety"
        assert len(result.counterfactual_hints) == 1
        assert result.counterfactual_hints[0].predicted_outcome == "DENY"

    def test_convert_without_counterfactuals(self):
        exp = _mock_explanation(include_cf=True)
        result = _convert_to_response(exp, include_counterfactuals=False)
        assert result.counterfactual_hints == []

    def test_convert_governance_dimension_string_fallback(self):
        exp = _mock_explanation()
        exp.factors = [_mock_factor(governance_dim_has_value=False)]
        result = _convert_to_response(exp, include_counterfactuals=False)
        assert result.factors[0].governance_dimension == "safety"

    def test_convert_counterfactual_outcome_string_fallback(self):
        exp = _mock_explanation()
        exp.counterfactual_hints = [_mock_counterfactual(outcome_has_value=False)]
        result = _convert_to_response(exp, include_counterfactuals=True)
        assert result.counterfactual_hints[0].predicted_outcome == "DENY"

    def test_convert_empty_factors_and_cf(self):
        exp = _mock_explanation(include_factors=False, include_cf=False)
        result = _convert_to_response(exp, include_counterfactuals=True)
        assert result.factors == []
        assert result.counterfactual_hints == []


# ---------------------------------------------------------------------------
# GET /{decision_id}/explain
# ---------------------------------------------------------------------------


class TestGetDecisionExplanation:
    """GET /api/v1/decisions/{decision_id}/explain"""

    def test_get_explanation_success(self, client: TestClient):
        mock_exp = _mock_explanation()
        mock_service = AsyncMock()
        mock_service.get_explanation = AsyncMock(return_value=mock_exp)

        mock_module = MagicMock()
        mock_module.ExplanationServiceAdapter = lambda *a, **kw: mock_service

        with patch.dict(
            "sys.modules",
            {
                "packages": MagicMock(),
                "packages.enhanced_agent_bus": MagicMock(),
                "packages.enhanced_agent_bus.facades": MagicMock(),
                "packages.enhanced_agent_bus.facades.agent_bus_facade": mock_module,
            },
        ):
            resp = client.get("/api/v1/decisions/dec-1/explain")
            assert resp.status_code == 200
            body = resp.json()
            assert body["decision_id"] == "dec-1"
            assert body["verdict"] == "ALLOW"

    def test_get_explanation_not_found(self, client: TestClient):
        mock_service = AsyncMock()
        mock_service.get_explanation = AsyncMock(return_value=None)

        with patch.dict(
            "sys.modules",
            {
                "packages": MagicMock(),
                "packages.enhanced_agent_bus": MagicMock(),
                "packages.enhanced_agent_bus.facades": MagicMock(),
                "packages.enhanced_agent_bus.facades.agent_bus_facade": MagicMock(
                    ExplanationServiceAdapter=lambda: mock_service,
                ),
            },
        ):
            resp = client.get("/api/v1/decisions/dec-1/explain")
            assert resp.status_code in (404, 500)

    def test_get_explanation_internal_error(self, client: TestClient):
        mock_service = AsyncMock()
        mock_service.get_explanation = AsyncMock(side_effect=RuntimeError("db down"))

        with patch.dict(
            "sys.modules",
            {
                "packages": MagicMock(),
                "packages.enhanced_agent_bus": MagicMock(),
                "packages.enhanced_agent_bus.facades": MagicMock(),
                "packages.enhanced_agent_bus.facades.agent_bus_facade": MagicMock(
                    ExplanationServiceAdapter=lambda: mock_service,
                ),
            },
        ):
            resp = client.get("/api/v1/decisions/dec-1/explain")
            assert resp.status_code == 500


# ---------------------------------------------------------------------------
# POST /explain
# ---------------------------------------------------------------------------


class TestGenerateDecisionExplanation:
    """POST /api/v1/decisions/explain"""

    def test_generate_explanation_success(self, client: TestClient):
        mock_exp = _mock_explanation()
        mock_service = AsyncMock()
        mock_service.generate_explanation = AsyncMock(return_value=mock_exp)

        mock_module = MagicMock()
        mock_module.ExplanationServiceAdapter = lambda *a, **kw: mock_service
        mock_module.ExplanationService = MagicMock()

        with patch.dict(
            "sys.modules",
            {
                "packages": MagicMock(),
                "packages.enhanced_agent_bus": MagicMock(),
                "packages.enhanced_agent_bus.facades": MagicMock(),
                "packages.enhanced_agent_bus.facades.agent_bus_facade": mock_module,
            },
        ):
            resp = client.post(
                "/api/v1/decisions/explain",
                json={
                    "message": {"content": "test action"},
                    "verdict": "ALLOW",
                    "context": {},
                    "include_counterfactuals": True,
                },
            )
            assert resp.status_code in (200, 500)

    def test_generate_explanation_error(self, client: TestClient):
        mock_module = MagicMock()
        mock_module.ExplanationServiceAdapter = MagicMock(side_effect=RuntimeError("fail"))
        mock_module.ExplanationService = MagicMock()

        with patch.dict(
            "sys.modules",
            {
                "packages": MagicMock(),
                "packages.enhanced_agent_bus": MagicMock(),
                "packages.enhanced_agent_bus.facades": MagicMock(),
                "packages.enhanced_agent_bus.facades.agent_bus_facade": mock_module,
            },
        ):
            resp = client.post(
                "/api/v1/decisions/explain",
                json={
                    "message": {"content": "test"},
                    "verdict": "DENY",
                },
            )
            assert resp.status_code == 500


# ---------------------------------------------------------------------------
# GET /governance-vector/schema
# ---------------------------------------------------------------------------


class TestGovernanceVectorSchema:
    """GET /api/v1/decisions/governance-vector/schema"""

    def test_schema_returns_200(self, client: TestClient):
        resp = client.get("/api/v1/decisions/governance-vector/schema")
        assert resp.status_code == 200
        body = resp.json()
        assert "dimensions" in body
        dims = body["dimensions"]
        assert set(dims.keys()) == {
            "safety",
            "security",
            "privacy",
            "fairness",
            "reliability",
            "transparency",
            "efficiency",
        }
        assert body["constitutional_hash"] == "608508a9bd224290"
        assert body["thresholds"]["escalation"] == 0.8

    def test_schema_dimension_structure(self, client: TestClient):
        resp = client.get("/api/v1/decisions/governance-vector/schema")
        body = resp.json()
        for dim in body["dimensions"].values():
            assert "description" in dim
            assert "range" in dim
            assert dim["range"] == [0.0, 1.0]
            assert "high_score_meaning" in dim
